"""Keeping track of mapping and reduce operations over tasks."""

from copy import deepcopy
import itertools
from math import prod
import logging
import typing as ty
from pydra.utils.typing import StateArray, TypeParser
from pydra.utils.general import ensure_list, attrs_values


logger = logging.getLogger("pydra")


OutputsType = ty.TypeVar("OutputsType")


class State:
    """
    A class that specifies a State of all tasks.

      * It's only used when a job have a splitter.
      * It contains all information about splitter, combiner, final splitter,
        and input values for specific job states
        (specified by the splitter and the input).
      * It also contains information about the final groups and the final splitter
        if combiner is available.

    Attributes
    ----------
    name : :obj:`str`
        name of the state that is the same as a name of the job
    splitter : :obj:`str`, :obj:`tuple`, :obj:`list`
        can be a str (name of a single input),
        tuple for scalar splitter, or list for outer splitter
    splitter_rpn_compact : :obj:`list`
        splitter in :abbr:`RPN (reverse Polish notation)`, using a compact
        notation for splitter from previous states, e.g. _NA
    splitter_rpn : :obj:`list`
        splitter represented in RPN,
        unwrapping splitters from previous states
    combiner : :obj:`list`
        list of fields that should be combined
        (order is not important)
    splitter_final :
        final splitter that includes the combining process
    other_states : :obj:`dict`
        used to create connections with previous states::

            {
              name of a previous state:
                (previous state, input from current state needed the connection)
            }

    inner_inputs : :obj:`dict`
        used to create connections with previous states
        ``{"{self.name}.input name for current inp": previous state}``
    states_ind : :obj:`list` of :obj:`dict`
        dictionary for every state that contains
        indices for all state inputs (i.e. inputs that are part of the splitter)
    states_val : :obj:`list` of :obj:`dict`
        dictionary for every state that contains
        values for all state inputs (i.e. inputs that are part of the splitter)
    inputs_ind : :obj:`list` of :obj:`dict`
        dictionary for every state that contains
        indices for all job inputs (i.e. inputs that are relevant
        for current job, can be outputs from previous nodes)
    group_for_inputs : :obj:`dict`
        specifying groups (axes) for each input field
        (depends on the splitter)
    group_for_inputs_final : :obj:`dict`
        specifying final groups (axes)
        for each input field (depends on the splitter and combiner)
    groups_stack_final : :obj:`list`
        specify stack of groups/axes (used to
        determine which field could be combined)
    final_combined_ind_mapping : :obj:`dict`
        mapping between final indices
        after combining and partial indices of the results

    """

    def __init__(
        self,
        name,
        splitter=None,
        combiner=None,
        container_ndim=None,
        other_states=None,
    ):
        """
        Initialize a state.

        Parameters
        ----------
        name : :obj:`str`
            name (should be the same as the job's name)
        splitter : :obj:`str`, or :obj:`tuple`, or :obj:`list`
            splitter of a job
        combiner : :obj:`str`, or :obj:`list`)
            field/fields used to combine results
        other_states :obj:`dict`:
            ``{name of a previous state: (prefious state,
            input from current state needed the connection)}``

        """
        self.name = name
        self.other_states = other_states
        self.splitter = splitter
        # temporary combiner
        self.combiner = combiner
        self.container_ndim = container_ndim or {}
        self._inner_container_ndim = {}
        self._inputs_ind = None
        # if other_states, the connections have to be updated
        if self.other_states:
            self.update_connections()

    def __str__(self):
        """Generate a string representation of the object."""
        return (
            f"State for {self.name} with a splitter: {self.splitter} "
            f"and combiner: {self.combiner}"
        )

    def depth(self, before_combine: bool = False) -> int:
        """Return the number of splits of the state, i.e. the number nested
        state arrays to wrap around the type of lazy out fields

        Parameters
        ----------
        before_combine : :obj:`bool`
            if True, the depth is after combining the fields, otherwise it is before
            any combinations

        Returns
        -------
        int
            number of splits in the state (i.e. linked splits only add 1)
        """

        # replace field names with 1 or 0 (1 if the field is included in the state)
        include_rpn = [
            (
                s
                if s in [".", "*"]
                else (1 if before_combine else int(s not in self.combiner))
            )
            for s in self.splitter_rpn
        ]

        stack = []
        for opr in include_rpn:
            if opr == ".":
                assert len(stack) >= 2
                opr1 = stack.pop()
                opr2 = stack.pop()
                stack.append(opr1 and opr2)
            elif opr == "*":
                assert len(stack) >= 2
                stack.append(stack.pop() + stack.pop())
            else:
                stack.append(opr)
        assert len(stack) == 1
        return stack[0]

    def nest_output_type(self, type_: type) -> type:
        """Nests a type of an output field in a combination of lists and state-arrays
        based on the state's splitter and combiner

        Parameters
        ----------
        type_ : type
            the type of the output field

        Returns
        -------
        type
            the nested type of the output field
        """

        state_array_depth = self.depth()

        # If there is a combination, it will get flattened into a single list
        if self.depth(before_combine=True) > state_array_depth:
            type_ = list[type_]

        # Nest the uncombined state arrays around the type
        for _ in range(state_array_depth):
            type_ = StateArray[type_]
        return type_

    @classmethod
    def combine_state_arrays(cls, type_: type) -> type:
        """Collapses (potentially nested) state array(s) into a single list"""
        if TypeParser.get_origin(type_) is StateArray:
            # Implicitly combine any remaining uncombined states into a single
            # list
            type_ = list[TypeParser.strip_splits(type_)[0]]
        return type_

    @property
    def splitter(self):
        """Get the splitter of the state."""
        return self._splitter

    @splitter.setter
    def splitter(self, splitter):
        if splitter and not isinstance(splitter, (str, tuple, list)):
            raise PydraStateError("splitter has to be a string, a tuple or a list")
        if splitter:
            self._splitter = add_name_splitter(splitter, self.name)
        else:
            self._splitter = None
        # updating splitter_rpn
        self._splitter_rpn_updates()

    def _splitter_rpn_updates(self):
        """updating splitter_rpn and splitter_rpn_compact"""
        try:
            self._splitter_rpn = splitter2rpn(
                self.splitter, other_states=self.other_states
            )
        # other_state might not be ready yet
        except PydraStateError:
            self._splitter_rpn = None

        if self.other_states or self._splitter_rpn is None:
            self._splitter_rpn_compact = splitter2rpn(
                self.splitter, other_states=self.other_states, state_fields=False
            )
        else:
            self._splitter_rpn_compact = self._splitter_rpn

    @property
    def splitter_rpn(self):
        """splitter in :abbr:`RPN (Reverse Polish Notation)`"""
        # if splitter_rpn was not calculated within splitter.setter
        if self._splitter_rpn is None:
            self._splitter_rpn = splitter2rpn(
                self.splitter, other_states=self.other_states
            )
        return self._splitter_rpn

    @property
    def splitter_rpn_compact(self):
        r"""splitter in :abbr:`RPN (Reverse Polish Notation)`
        with a compact representation of the prev-state part (i.e. without unwrapping
        the part that comes from the previous states), e.g., [_NA, _NB, \*]
        """
        return self._splitter_rpn_compact

    @property
    def splitter_final(self):
        """the final splitter, after removing the combined fields"""
        return rpn2splitter(self.splitter_rpn_final)

    @property
    def splitter_rpn_final(self):
        if self.combiner:
            _splitter_rpn_final = remove_inp_from_splitter_rpn(
                deepcopy(self.splitter_rpn),
                self.current_combiner_all + self.prev_state_combiner_all,
            )
            return _splitter_rpn_final
        else:
            return self.splitter_rpn

    @property
    def current_splitter(self):
        """the current part of the splitter,
        i.e. the part that is related to the current job's state only
        (doesn't include fields propagated from the previous tasks)
        """
        if hasattr(self, "_current_splitter"):
            return self._current_splitter
        else:
            return self.splitter

    @property
    def inputs_ind(self):
        """dictionary for every state that contains indices for all job inputs
        (i.e. inputs that are relevant for current job, can be outputs from previous nodes)
        """
        if self._inputs_ind is None:
            raise RuntimeError(
                "inputs_ind is not set, please run prepare_states() on the state first"
            )
        return self._inputs_ind

    @current_splitter.setter
    def current_splitter(self, current_splitter):
        self._current_splitter = current_splitter
        # updating rpn
        self._current_splitter_rpn_updates()

    def _current_splitter_rpn_updates(self):
        """updating current_splitter_rpn"""
        if self._current_splitter:
            self._current_splitter_rpn = splitter2rpn(
                self.current_splitter, other_states=self.other_states
            )
        else:
            self._current_splitter_rpn = []

    @property
    def current_splitter_rpn(self):
        """the current part of the splitter using RPN"""
        if hasattr(self, "_current_splitter_rpn"):
            return self._current_splitter_rpn
        else:
            return self.splitter_rpn

    @property
    def prev_state_splitter(self):
        """the prev-state part of the splitter,
        i.e. the part that comes from the previous tasks' states
        """
        if hasattr(self, "_prev_state_splitter"):
            return self._prev_state_splitter
        else:
            return None

    @prev_state_splitter.setter
    def prev_state_splitter(self, prev_state_splitter):
        self._prev_state_splitter = prev_state_splitter
        # updating rpn splitters
        self._prev_state_splitter_rpn_updates()

    def _prev_state_splitter_rpn_updates(self):
        """updating prev_state_splitter_rpn/_rpn_compact"""
        if self._prev_state_splitter:
            self._prev_state_splitter_rpn = splitter2rpn(
                self.prev_state_splitter, other_states=self.other_states
            )
        else:
            self._prev_state_splitter_rpn = []

        if self.other_states:
            self._prev_state_splitter_rpn_compact = splitter2rpn(
                self.prev_state_splitter,
                other_states=self.other_states,
                state_fields=False,
            )
        else:
            self._prev_state_splitter_rpn_compact = self._prev_state_splitter_rpn

    @property
    def prev_state_splitter_rpn(self):
        """the prev-state art of the splitter using RPN"""
        return self._prev_state_splitter_rpn

    @property
    def prev_state_splitter_rpn_compact(self):
        r"""the prev-state part of the splitter using RPN in a compact form,
        (without unwrapping the states from previous nodes), e.g. [_NA, _NB, \*]
        """
        return self._prev_state_splitter_rpn_compact

    @property
    def container_ndim_all(self):
        # adding inner_container_ndim to the general container_dimension provided by the users
        container_ndim_all = deepcopy(self.container_ndim)
        for k, v in self._inner_container_ndim.items():
            container_ndim_all[k] = container_ndim_all.get(k, 1) + v
        return container_ndim_all

    @property
    def combiner(self):
        """the combiner associated to the state."""
        return self._combiner

    @combiner.setter
    def combiner(self, combiner):
        if combiner:
            if not isinstance(combiner, (str, list)):
                raise PydraStateError("combiner has to be a string or a list")
            self._combiner = add_name_combiner(ensure_list(combiner), self.name)
        else:
            self._combiner = []

    @property
    def current_combiner(self):
        """the current part of the combiner,
        i.e. the part that is related to the current job's state only
        (doesn't include fields propagated from the previous tasks)
        """
        return [comb for comb in self.combiner if self.name in comb]

    @property
    def current_combiner_all(self):
        """the current part of the combiner including all the fields
        that should be combined (i.e. not only the fields that are explicitly
        set, but also the fields that re in the same group/axis and had to be combined
        together, e.g., if splitter is (a, b) a and b has to be combined together)
        """
        if hasattr(self, "_current_combiner_all"):
            return self._current_combiner_all
        else:
            return self.current_combiner

    @property
    def prev_state_combiner(self):
        """the prev-state part of the combiner,
        i.e. the part that comes from the previous tasks' states
        """
        if hasattr(self, "_prev_state_combiner"):
            return self._prev_state_combiner
        else:
            return list(set(self.combiner) - set(self.current_combiner))

    @property
    def prev_state_combiner_all(self):
        """the prev-state part of the combiner including all the fields
        that should be combined (i.e. not only the fields that are explicitly
        set, but also the fields that re in the same group/axis and had to be combined
        together, e.g., if splitter is (a, b) a and b has to be combined together)
        """
        if hasattr(self, "_prev_state_combiner_all"):
            return list(set(self._prev_state_combiner_all))
        else:
            return self.prev_state_combiner

    @property
    def other_states(self):
        """specifies the connections with previous states, uses dictionary:
        {name of a previous state: (previous state, input field from current state)}
        """
        return self._other_states

    @other_states.setter
    def other_states(self, other_states):
        if other_states:
            if not isinstance(other_states, dict):
                raise PydraStateError("other states has to be a dictionary")
            else:
                for key, val in other_states.items():
                    if not val:
                        raise PydraStateError(f"connection from node {key} is empty")
            # ensuring that the connected fields are set as a list
            self._other_states = {
                nm: (st, ensure_list(flds)) for nm, (st, flds) in other_states.items()
            }
        else:
            self._other_states = {}

    @property
    def inner_inputs(self):
        """specifies connections between fields from the current state
        with the specific state from the previous states, uses dictionary
        ``{input name for current state: the previous state}``
        """
        if self.other_states:
            _inner_inputs = {}
            for name, (st, inp_l) in self.other_states.items():
                if f"_{st.name}" in self.splitter_rpn_compact:
                    for inp in inp_l:
                        _inner_inputs[f"{self.name}.{inp}"] = st
            return _inner_inputs
        else:
            return {}

    def update_connections(self, new_other_states=None, new_combiner=None):
        """updating connections, can use a new other_states and combiner

        Parameters
        ----------
        new_other_states : :obj:`dict`, optional
            dictionary with new other_states, will be set before updating connections
        new_combiner : :obj:`str`, or :obj:`list`, optional
            new combiner
        """
        if new_other_states:
            self.other_states = new_other_states
        self._connect_splitters()
        if new_combiner:
            self.combiner = new_combiner

    def _connect_splitters(self):
        """
        Connect splitters from the previous nodes.
        Evaluates the prev-state part of the splitter (i.e. the part from the previous states)
        and the current part of the splitter(i.e., the current state).
        If the prev-state splitter is not provided the splitter has to be completed.
        """
        # TODO: should this be in the prev-state_Splitter property?
        if self.splitter:
            # if splitter is string, have to check if this is prev-state
            # or current part (prev-state is required)
            if isinstance(self.splitter, str):
                # so this is the prev-state part
                if self.splitter.startswith("_"):
                    self.prev_state_splitter = self._complete_prev_state(
                        prev_state=self.splitter
                    )
                    self.current_splitter = None
                else:  # this is the current part
                    self.prev_state_splitter = self._complete_prev_state()
                    self.current_splitter = self.splitter
            elif isinstance(self.splitter, (tuple, list)):
                lr_flag = self._prevst_current_check(self.splitter)
                if lr_flag == "prev-state":
                    self.prev_state_splitter = self._complete_prev_state(
                        prev_state=self.splitter
                    )
                    self.current_splitter = None
                elif lr_flag == "current":
                    self.prev_state_splitter = self._complete_prev_state()
                    self.current_splitter = self.splitter
                elif lr_flag == "[prev-state, current]":
                    self.prev_state_splitter = self._complete_prev_state(
                        prev_state=self.splitter[0]
                    )
                    self.current_splitter = self.splitter[1]
        else:
            # if there is no splitter, I create the prev-state part
            self.prev_state_splitter = self._complete_prev_state()
            self.current_splitter = None

        if self.current_splitter:
            self.splitter = [
                deepcopy(self.prev_state_splitter),
                deepcopy(self.current_splitter),
            ]
        else:
            self.splitter = deepcopy(self.prev_state_splitter)

    def _complete_prev_state(self, prev_state=None):
        """Add all splitters from the previous nodes (completing the prev-state part).

        Parameters
        ----------
        prev_state :  :obj:`str`, or :obj:`list`, or :obj:`tuple`, optional
            the prev-state part of the splitter, that has to be completed
        """
        if prev_state:
            rpn_prev_state = splitter2rpn(
                prev_state, other_states=self.other_states, state_fields=False
            )
            for name, (st, inp) in list(self.other_states.items())[::-1]:
                if f"_{name}" not in rpn_prev_state and st.splitter_final:
                    prev_state = [f"_{name}", prev_state]
        else:
            prev_state = [f"_{name}" for name in self.other_states]
        # TODO: what if prev state is a tuple
        if isinstance(prev_state, list):
            prev_state = self._remove_repeated(prev_state)
            prev_state = self._add_state_history(ensure_list(prev_state))
        if len(prev_state) == 1:
            prev_state = prev_state[0]
        return prev_state

    def _remove_repeated(self, previous_splitters):
        """removing states from previous tasks that are repeated"""
        for el in previous_splitters:
            if el[1:] not in self.other_states:
                raise PydraStateError(
                    f"can't ask for splitter from {el[1:]}, other nodes that are connected: "
                    f"{self.other_states}"
                )

        repeated = {
            (el, previous_splitters.count(el))
            for el in previous_splitters
            if previous_splitters.count(el) > 1
        }
        if repeated:
            # assuming that I want to remove from right
            previous_splitters.reverse()
            for el, cnt in repeated:
                for ii in range(cnt):
                    previous_splitters.remove(el)
            previous_splitters.reverse()
        return previous_splitters

    def _add_state_history(self, previous_splitters):
        """analysing history of each state from previous states
        expanding previous_splitters list and oter_states
        """
        othst_w_currst = []  # el from other_states that have only current states
        othst_w_prevst = (
            []
        )  # el from other_states that have only prev st (nm, st.prev_st_spl)
        othst_w_currst_prevst = (
            []
        )  # el from other_states that have both (nm, st.prev_st_spl)
        for el in previous_splitters:
            nm = el[1:]
            st = self.other_states[nm][0]
            if not st.other_states:
                # states that has no other connections
                othst_w_currst.append(el)
            else:  # element has previous_connection
                if st.current_splitter:  # final?
                    # states that has previous connections and it's own splitter
                    othst_w_currst_prevst.append((el, st.prev_state_splitter))
                else:
                    # states with previous connections but no additional splitter
                    othst_w_prevst.append((el, st.prev_state_splitter))

        for el in othst_w_prevst:
            el_nm, el_prevst = el[0][1:], el[1]
            # checking if the element's previous state is not already included
            repeated_prev = set(ensure_list(el_prevst)).intersection(othst_w_currst)
            if repeated_prev:
                for r_el in repeated_prev:
                    r_nm = r_el[1:]
                    # updating self.other_states
                    self._other_states[r_nm] = (
                        self.other_states[r_nm][0],
                        self.other_states[r_nm][1] + self.other_states[el_nm][1],
                    )
                new_st = set(ensure_list(el_prevst)) - set(othst_w_currst)
                if not new_st:
                    # removing element from the previous_splitter if no new_st
                    previous_splitters.remove(el[0])
                else:
                    # adding elements to self.other_states
                    for new_el in new_st:
                        new_nm = new_el[1:]
                        self._other_states[new_nm] = (
                            self.other_states[el_nm][0].other_states[new_nm][0],
                            self.other_states[el_nm][1],
                        )
                    # removing el of the splitter and adding new_st instead
                    ind = previous_splitters.index(el[0])
                    if ind == len(previous_splitters) - 1:
                        previous_splitters = previous_splitters[:-1] + list(new_st)
                    else:
                        previous_splitters = (
                            previous_splitters[:ind]
                            + list(new_st)
                            + previous_splitters[ind + 1 :]
                        )
        # TODO: this part is not tested, needs more work
        for el in othst_w_currst_prevst:
            repeated_prev = set(ensure_list(el[1])).intersection(othst_w_currst)
            # removing el if it's repeated: in el.other_states and othst_w_currst
            if repeated_prev:
                for r_el in repeated_prev:
                    previous_splitters.remove(r_el)
        return previous_splitters

    def _prevst_current_check(self, splitter_part, check_nested=True):
        """
        Check if splitter_part is purely prev-state part, the current part,
        or mixed ([prev-state, current]) if the splitter_part is a list (outer splitter)

        Parameters
        ----------
        splitter_part : :obj:`str`, or :obj:`list`, or :obj:`tuple`
            Part of the splitter that is being check
        check_nested : :obj:`bool`, optional
            If True, the nested parts are checked.

        Returns
        -------
        str
           describes the type - "prev-state" or "current"

        If the splitter_part is mixed exception is raised.

        """
        rpn_part = splitter2rpn(
            splitter_part, other_states=self.other_states, state_fields=False
        )
        inputs_in_splitter = [i for i in rpn_part if i not in ["*", "."]]
        others_in_splitter = [
            True if el.startswith("_") else False for el in inputs_in_splitter
        ]
        if all(others_in_splitter):
            return "prev-state"
        elif (not all(others_in_splitter)) and (not any(others_in_splitter)):
            return "current"
        elif (
            isinstance(self.splitter, list)
            and check_nested
            and self._prevst_current_check(self.splitter[0], check_nested=False)
            == "prev-state"
            and self._prevst_current_check(self.splitter[1], check_nested=False)
            == "current"
        ):
            # the prev-state and the current parts separated in outer scalar
            return "[prev-state, current]"
        else:
            raise PydraStateError(
                "prev-state and current splitters are mixed - splitter invalid"
            )

    def set_input_groups(self, state_fields=True):
        """Evaluates groups, especially the final groups that address the combiner.

        Parameters
        ----------
        state_fields : :obj:`bool`
            if False the splitter from the previous states are unwrapped
        """
        current_splitter_rpn = splitter2rpn(
            self.current_splitter,
            other_states=self.other_states,
            state_fields=state_fields,
        )
        # merging groups from previous nodes if any input come from previous states
        if self.other_states:
            self._merge_previous_groups()
        keys_f, group_for_inputs_f, groups_stack_f, combiner_all = splits_groups(
            current_splitter_rpn,
            combiner=self.current_combiner,
            inner_inputs=self.inner_inputs,
        )
        self._current_combiner_all = combiner_all
        if (
            self.prev_state_splitter and state_fields
        ):  # if splitter has also the prev-state part
            self._current_keys_final = keys_f
            self._current_group_for_inputs_final = group_for_inputs_f
            self._current_groups_stack_final = groups_stack_f
            if (
                self.current_splitter
            ):  # if the current part, adding groups from current st
                self._add_current_groups()

        else:
            self.group_for_inputs_final = group_for_inputs_f
            self.groups_stack_final = groups_stack_f
            self.keys_final = keys_f

    def _merge_previous_groups(self):
        """Merge groups from  all previous nodes."""
        last_gr = 0
        self.groups_stack_final = []
        self.group_for_inputs_final = {}
        self.keys_final = []
        if self.prev_state_combiner:
            _, _, _, self._prev_state_combiner_all = splits_groups(
                self.prev_state_splitter_rpn, combiner=self.prev_state_combiner
            )
        for i, prev_nm in enumerate(self.prev_state_splitter_rpn_compact):
            if prev_nm in ["*", "."]:
                continue
            if (
                i + 1 < len(self.prev_state_splitter_rpn_compact)
                and self.prev_state_splitter_rpn_compact[i + 1] == "."
            ):
                last_gr = last_gr - 1
            if prev_nm[1:] not in self.other_states:
                raise PydraStateError(
                    f"can't ask for splitter from {prev_nm[1:]}, "
                    f"other nodes that are connected: {self.other_states}"
                )
            st = self.other_states[prev_nm[1:]][0]
            # checking if prev-state combiner contains any element from the st splitter
            st_combiner = [
                comb
                for comb in self.prev_state_combiner_all
                if comb in st.splitter_rpn_final
            ]
            if not hasattr(st, "keys_final"):
                st.set_input_groups()
            if st_combiner:
                # keys and groups from previous states
                # after taking into account combiner from current state
                (
                    keys_f_st,
                    group_for_inputs_f_st,
                    groups_stack_f_st,
                    combiner_all_st,
                ) = splits_groups(
                    st.splitter_rpn_final,
                    combiner=st_combiner,
                    inner_inputs=st.inner_inputs,
                )
                self.keys_final += keys_f_st  # st.keys_final
                if not hasattr(st, "group_for_inputs_final"):
                    raise PydraStateError("previous state has to run first")
                group_for_inputs = group_for_inputs_f_st
                groups_stack = groups_stack_f_st
                self._prev_state_combiner_all += combiner_all_st
            else:
                # if no element from st.splitter is in the current combiner,
                # using st attributes without changes
                if st.keys_final:
                    self.keys_final += st.keys_final
                group_for_inputs = st.group_for_inputs_final
                groups_stack = st.groups_stack_final

            group_for_inputs = {k: v + last_gr for k, v in group_for_inputs.items()}
            self.group_for_inputs_final.update(group_for_inputs)

            nmb_gr = 0
            for i, groups in enumerate(groups_stack):
                if i < len(self.groups_stack_final):
                    for gr in groups:
                        nmb_gr += 1
                        if gr + last_gr not in self.groups_stack_final[i]:
                            self.groups_stack_final[i].append(gr + last_gr)
                else:
                    self.groups_stack_final.append([gr + last_gr for gr in groups])
                    nmb_gr += len(groups)
            last_gr += nmb_gr

    def _add_current_groups(self):
        """Add additional groups from the current state."""
        self.keys_final += self._current_keys_final
        nr_gr_f = max(self.group_for_inputs_final.values()) + 1
        for inp, grs in self._current_group_for_inputs_final.items():
            if isinstance(grs, int):
                grs = grs + nr_gr_f
            else:  # a list
                grs = [gr + nr_gr_f for gr in grs]
            self.group_for_inputs_final[inp] = grs
        for i, stack in enumerate(self._current_groups_stack_final):
            if i == 0:
                stack = [gr + nr_gr_f for gr in stack]
                self.groups_stack_final[-1] += stack
            else:
                stack = [gr + nr_gr_f for gr in stack]
                self.groups_stack_final.append(stack)

    def splitter_validation(self):
        """validating if the splitter is correct (after all states are connected)"""
        for spl in self.splitter_rpn_compact:
            if not (
                spl in [".", "*"]
                or (spl.startswith("_") and spl[1:] in self.other_states)
                or spl.split(".")[0] == self.name
            ):
                raise PydraStateError(
                    "can't include {} in the splitter, consider using _{}".format(
                        spl, spl.split(".")[0]
                    )
                )

    def combiner_validation(self):
        """validating if the combiner is correct (after all states are connected)"""
        if local_names := set(
            c for c in self.combiner if c.startswith(self.name + ".")
        ):
            if not self.splitter:
                raise PydraStateError(
                    "splitter has to be set before setting combiner with field names "
                    f"in the current node {list(local_names)}"
                )
            if missing := local_names - set(self.splitter_rpn):
                raise PydraStateError(
                    "The following field names from the current node referenced in the "
                    f"combiner, {list(missing)} are not in the splitter"
                )

    def prepare_states(
        self,
        inputs: dict[str, ty.Any],
        container_ndim: dict[str, int] | None = None,
    ):
        """
        Prepare a full list of state indices and state values.

        State Indices
            number of elements depends on the splitter

        State Values
            specific elements from inputs that can be used running interfaces
        """
        # checking if splitter and combiner have valid forms
        self.splitter_validation()
        self.combiner_validation()
        self.set_input_groups()
        self.inputs = inputs
        if container_ndim is not None:
            self.container_ndim = container_ndim
        if self.other_states:
            st: State
            for nm, (st, _) in self.other_states.items():
                self.inputs.update(st.inputs)
                self.container_ndim.update(st.container_ndim_all)

        self.prepare_states_ind()
        self.prepare_states_val()

    def prepare_states_ind(self):
        """
        Calculate a list of dictionaries with state indices.

        Uses splits.

        """
        # removing elements that are connected to inner splitter
        # (they will be taken into account in splits anyway)
        # _comb part will be used in prepare_states_combined_ind
        # TODO: need tests in test_Workflow.py
        elements_to_remove = []
        elements_to_remove_comb = []
        for name, (st, inp_l) in self.other_states.items():
            for inp in inp_l:
                if (
                    f"{self.name}.{inp}" in self.splitter_rpn
                    and f"_{name}" in self.splitter_rpn_compact
                ):
                    elements_to_remove.append(f"_{name}")
                    if f"{self.name}.{inp}" not in self.combiner:
                        elements_to_remove_comb.append(f"_{name}")

        partial_rpn = remove_inp_from_splitter_rpn(
            deepcopy(self.splitter_rpn_compact), elements_to_remove
        )
        values_out_pr, keys_out_pr = self.splits(
            partial_rpn,
        )
        values_pr = list(values_out_pr)

        self.ind_l = values_pr
        self.keys = keys_out_pr
        self.states_ind = list(iter_splits(values_pr, self.keys))
        self.keys_final = self.keys
        if self.combiner:
            self.prepare_states_combined_ind(elements_to_remove_comb)
        else:
            self.ind_l_final = self.ind_l
            self.keys_final = self.keys
            self.final_combined_ind_mapping = {
                i: [i] for i in range(len(self.states_ind))
            }
            self.states_ind_final = self.states_ind
        return self.states_ind

    def prepare_states_combined_ind(self, elements_to_remove_comb):
        """Prepare the final list of dictionaries with indices after combiner.

        Parameters
        ----------
        elements_to_remove_comb : :obj:`list`
            elements of the splitter that should be removed due to the combining
        """
        partial_rpn_compact = remove_inp_from_splitter_rpn(
            deepcopy(self.splitter_rpn_compact), elements_to_remove_comb
        )
        # combiner can have parts from the prev-state splitter, so have to have rpn with states
        partial_rpn = splitter2rpn(
            rpn2splitter(partial_rpn_compact), other_states=self.other_states
        )
        combined_rpn = remove_inp_from_splitter_rpn(
            deepcopy(partial_rpn),
            self.current_combiner_all + self.prev_state_combiner_all,
        )
        if combined_rpn:
            val_r, key_r = self.splits(
                combined_rpn,
            )
            values = list(val_r)
        else:
            values = []
            key_r = []

        keys_out = key_r
        if values:
            self.ind_l_final = values
            self.keys_final = keys_out
            # groups after combiner
            ind_map = {
                tuple(flatten(tup, max_depth=10)): ind
                for ind, tup in enumerate(self.ind_l_final)
            }
            self.final_combined_ind_mapping = {
                i: [] for i in range(len(self.ind_l_final))
            }
            for ii, st in enumerate(self.states_ind):
                ind_f = tuple([st[k] for k in self.keys_final])
                self.final_combined_ind_mapping[ind_map[ind_f]].append(ii)
        else:
            self.ind_l_final = values
            self.keys_final = keys_out
            # should be 0 or None?
            self.final_combined_ind_mapping = {0: list(range(len(self.states_ind)))}
        self.states_ind_final = list(iter_splits(self.ind_l_final, self.keys_final))

    def prepare_states_val(self):
        """Evaluate states values having states indices."""
        self.states_val = list(
            map_splits(
                self.states_ind, self.inputs, container_ndim=self.container_ndim_all
            )
        )
        return self.states_val

    def prepare_inputs(self):
        """
        Preparing inputs indices, merges input from previous states.

        Includes indices for fields from inner splitters
        (removes elements connected to the inner splitters fields).

        """
        if not self.other_states:
            self._inputs_ind = self.states_ind
        else:
            # elements from the current node (the current part of the splitter)
            if self.current_splitter_rpn:
                values_inp, keys_inp = self.splits(
                    self.current_splitter_rpn,
                )
                inputs_ind = values_inp
            else:
                keys_inp = []
                inputs_ind = []

            # merging elements that come from previous nodes outputs
            # states that are connected to inner splitters are treated differently
            # (already included in inputs_ind)
            keys_inp_prev = []
            inputs_ind_prev = []
            connected_to_inner = []
            for ii, el in enumerate(self.prev_state_splitter_rpn_compact):
                if el in ["*", "."]:
                    continue
                st, inp_l = self.other_states[el[1:]]
                inp_l = [f"{self.name}.{inp}" for inp in inp_l]
                if set(inp_l).intersection(self.splitter_rpn):  # inner splitter
                    connected_to_inner += [
                        el for el in st.splitter_rpn_final if el not in [".", "*"]
                    ]
                else:  # previous states that are not connected to inner splitter
                    st_ind = range(len(st.states_ind_final))
                    if inputs_ind_prev:
                        # in case the prev-state part has scalar parts (not very well tested)
                        if self.prev_state_splitter_rpn_compact[ii + 1] == ".":
                            inputs_ind_prev = op["."](inputs_ind_prev, st_ind)
                        else:
                            inputs_ind_prev = op["*"](inputs_ind_prev, st_ind)
                    else:
                        # TODO: more tests needed
                        inputs_ind_prev = op["."](*[st_ind] * len(inp_l))
                    keys_inp_prev += inp_l
            keys_inp = keys_inp_prev + keys_inp

            if inputs_ind and inputs_ind_prev:
                inputs_ind = op["*"](inputs_ind_prev, inputs_ind)
            elif inputs_ind:
                inputs_ind = op["*"](inputs_ind)
            elif inputs_ind_prev:
                inputs_ind = op["*"](inputs_ind_prev)
            else:
                inputs_ind = []

            # iter_splits using inputs from current state/node
            self._inputs_ind = list(iter_splits(inputs_ind, keys_inp))
            # removing elements that are connected to inner splitter
            # TODO - add tests to test_workflow.py (not sure if we want to remove it)
            for el in connected_to_inner:
                [dict.pop(el) for dict in self._inputs_ind]

    def splits(self, splitter_rpn):
        """
        Splits input variable as specified by splitter

        Parameters
        ----------
        splitter_rpn : list
            splitter in RPN notation
        Returns
        -------
        splitter : list
            each element contains indices for input variables
        keys: list
            names of input variables

        """
        # analysing states from connected tasks if inner_inputs
        previous_states_ind = {
            f"_{v.name}": (v.ind_l_final, v.keys_final)
            for v in self.inner_inputs.values()
        }

        # when splitter is a single element (no operators)
        if len(splitter_rpn) == 1:
            op_single = splitter_rpn[0]
            # splitter comes from the previous state
            if op_single.startswith("_"):
                return previous_states_ind[op_single]
            return self._single_op_splits(op_single)

        stack = []
        keys = []

        # iterating splitter_rpn
        for token in splitter_rpn:
            if token not in [".", "*"]:  # token is one of the input var
                # adding variable to the stack
                stack.append(token)
            else:
                # removing Right and Left var from the stack
                term_R = stack.pop()
                term_L = stack.pop()

                # analysing and processing Left and Right terms
                # both terms (Left and Right) are strings, so they were not processed yet by the function
                if isinstance(term_L, str) and isinstance(term_R, str):
                    shape_L, var_ind_L, new_keys_L = self._processing_terms(
                        term_L, previous_states_ind
                    )
                    shape_R, var_ind_R, new_keys_R = self._processing_terms(
                        term_R, previous_states_ind
                    )
                    keys = keys + new_keys_L + new_keys_R
                elif isinstance(term_L, str):
                    shape_L, var_ind_L, new_keys_L = self._processing_terms(
                        term_L, previous_states_ind
                    )
                    var_ind_R, shape_R = term_R
                    keys = new_keys_L + keys
                elif isinstance(term_R, str):
                    shape_R, var_ind_R, new_keys_R = self._processing_terms(
                        term_R, previous_states_ind
                    )
                    var_ind_L, shape_L = term_L
                    keys = keys + new_keys_R
                else:
                    var_ind_L, shape_L = term_L
                    var_ind_R, shape_R = term_R

                # checking shapes and creating newshape
                if token == ".":
                    if shape_L != shape_R:
                        raise ValueError(
                            f"Operands {term_R} and {term_L} do not have same shape"
                        )
                    newshape = shape_R
                elif token == "*":
                    newshape = tuple(list(shape_L) + list(shape_R))

                # creating a new iterator with all indices for the current operation
                # and pushing it to the stack
                pushval = (op[token](var_ind_L, var_ind_R), newshape)
                stack.append(pushval)

        # when everything is processed it should be one element in the stack
        # that contains iterator with variable indices after splitting for all keys
        var_ind = stack.pop()
        if isinstance(var_ind, tuple):
            var_ind = var_ind[0]

        return var_ind, keys

    def _processing_terms(self, term, previous_states_ind):
        """processing a specific term to get new keys from the term,
        an iterator with variable indices and matching keys
        """
        if term.startswith("_"):
            var_ind, new_keys = previous_states_ind[term]
            shape = (len(var_ind),)
        else:
            container_ndim = self.container_ndim_all.get(term, 1)
            shape = input_shape(self.inputs[term], container_ndim=container_ndim)
            var_ind = range(prod(shape))
            new_keys = [term]
            # checking if the term is in inner_inputs
            if term in self.inner_inputs:
                # TODO: have to be changed if differ length
                inner_len = [shape[-1]] * prod(shape[:-1])
                # this come from the previous node
                outer_ind = self.inner_inputs[term].ind_l
                var_ind_out = itertools.chain.from_iterable(
                    itertools.repeat(x, n) for x, n in zip(outer_ind, inner_len)
                )
                var_ind = op["."](var_ind_out, var_ind)
                new_keys = self.inner_inputs[term].keys_final + new_keys

        return shape, var_ind, new_keys

    def _single_op_splits(self, op_single):
        """splits function if splitter is a singleton"""
        shape = input_shape(
            self.inputs[op_single],
            container_ndim=self.container_ndim_all.get(op_single, 1),
        )
        val_ind = range(prod(shape))
        if op_single in self.inner_inputs:
            # TODO: have to be changed if differ length
            inner_len = [shape[-1]] * prod(shape[:-1])

            # this come from the previous node
            outer_ind = self.inner_inputs[op_single].ind_l
            op_out = itertools.chain.from_iterable(
                itertools.repeat(x, n) for x, n in zip(outer_ind, inner_len)
            )
            res = op["."](op_out, val_ind)
            val = res
            keys = self.inner_inputs[op_single].keys_final + [op_single]
            return val, keys
        else:
            val = op["*"](val_ind)
            keys = [op_single]
            return val, keys

    def _get_element(self, value: ty.Any, field_name: str, ind: int) -> ty.Any:
        """
        Extracting element of the inputs taking into account
        container dimension of the specific element that can be set in self.state.container_ndim.
        If input name is not in container_ndim, it is assumed that the input values has
        a container dimension of 1, so only the most outer dim will be used for splitting.

        Parameters
        ----------
        value : Any
            inputs of the job
        field_name : str
            name of the input field
        ind : int
            index of the element

        Returns
        -------
        Any
            specific element of the input field
        """
        if f"{self.name}.{field_name}" in self.container_ndim_all:
            return list(
                flatten(
                    ensure_list(value),
                    max_depth=self.container_ndim_all[f"{self.name}.{field_name}"],
                )
            )[ind]
        else:
            return value[ind]


def splitter2rpn(splitter, other_states=None, state_fields=True):
    """
    Translate user-provided splitter into *reverse polish notation*.

    The reverse polish notation is imposed by :class:`~pydra.engine.state.State`.

    Parameters
    ----------
    splitter :
        splitter (standard form)
    other_states :
        other states that are connected to the state
    state_fields : :obj:`bool`
        if False the splitter from the previous states are unwrapped

    """
    if not splitter:
        return []
    output_splitter = []
    _ordering(
        deepcopy(splitter),
        i=0,
        output_splitter=output_splitter,
        other_states=deepcopy(other_states),
        state_fields=state_fields,
    )
    return output_splitter


def _ordering(
    el, i, output_splitter, current_sign=None, other_states=None, state_fields=True
):
    """Get a proper order of fields and signs (used by splitter2rpn)."""
    if type(el) is tuple:
        # checking if the splitter dont contain splitter from previous nodes
        # i.e. has str "_NA", etc.
        if len(el) == 1:
            # treats .split(("x",)) like .split("x")
            el = el[0]
            _ordering(el, i, output_splitter, current_sign, other_states, state_fields)
        else:
            if type(el[0]) is str and el[0].startswith("_"):
                node_nm = el[0][1:]
                if node_nm not in other_states and state_fields:
                    raise PydraStateError(
                        "can't ask for splitter from {}, other nodes that are connected: {}".format(
                            node_nm, other_states.keys()
                        )
                    )
                elif state_fields:
                    splitter_mod = add_name_splitter(
                        splitter=other_states[node_nm][0].splitter_final, name=node_nm
                    )
                    el = (splitter_mod, el[1])
                    if other_states[node_nm][0].other_states:
                        other_states.update(other_states[node_nm][0].other_states)
            if type(el[1]) is str and el[1].startswith("_"):
                node_nm = el[1][1:]
                if node_nm not in other_states and state_fields:
                    raise PydraStateError(
                        "can't ask for splitter from {}, other nodes that are connected: {}".format(
                            node_nm, other_states.keys()
                        )
                    )
                elif state_fields:
                    splitter_mod = add_name_splitter(
                        splitter=other_states[node_nm][0].splitter_final, name=node_nm
                    )
                    el = (el[0], splitter_mod)
                    if other_states[node_nm][0].other_states:
                        other_states.update(other_states[node_nm][0].other_states)
            _iterate_list(
                el,
                ".",
                other_states,
                output_splitter=output_splitter,
                state_fields=state_fields,
            )
    elif type(el) is list:
        if len(el) == 1:
            # treats .split(["x"]) like .split("x")
            el = el[0]
            _ordering(el, i, output_splitter, current_sign, other_states, state_fields)
        else:
            if type(el[0]) is str and el[0].startswith("_"):
                node_nm = el[0][1:]
                if node_nm not in other_states and state_fields:
                    raise PydraStateError(
                        "can't ask for splitter from {}, other nodes that are connected: {}".format(
                            node_nm, other_states.keys()
                        )
                    )
                elif state_fields:
                    splitter_mod = add_name_splitter(
                        splitter=other_states[node_nm][0].splitter_final, name=node_nm
                    )
                    el[0] = splitter_mod
                    if other_states[node_nm][0].other_states:
                        other_states.update(other_states[node_nm][0].other_states)
            if type(el[1]) is str and el[1].startswith("_"):
                node_nm = el[1][1:]
                if node_nm not in other_states and state_fields:
                    raise PydraStateError(
                        "can't ask for splitter from {}, other nodes that are connected: {}".format(
                            node_nm, other_states.keys()
                        )
                    )
                elif state_fields:
                    splitter_mod = add_name_splitter(
                        splitter=other_states[node_nm][0].splitter_final, name=node_nm
                    )
                    el[1] = splitter_mod
                    if other_states[node_nm][0].other_states:
                        other_states.update(other_states[node_nm][0].other_states)
            _iterate_list(
                el,
                "*",
                other_states,
                output_splitter=output_splitter,
                state_fields=state_fields,
            )
    elif type(el) is str:
        if el.startswith("_"):
            node_nm = el[1:]
            if node_nm not in other_states and state_fields:
                raise PydraStateError(
                    "can't ask for splitter from {}, other nodes that are connected: {}".format(
                        node_nm, other_states.keys()
                    )
                )
            elif state_fields:
                splitter_mod = add_name_splitter(
                    splitter=other_states[node_nm][0].splitter_final, name=node_nm
                )
                el = splitter_mod
                if other_states[node_nm][0].other_states:
                    other_states.update(other_states[node_nm][0].other_states)
        if type(el) is str:
            output_splitter.append(el)
        elif type(el) is tuple:
            _iterate_list(
                el,
                ".",
                other_states,
                output_splitter=output_splitter,
                state_fields=state_fields,
            )
        elif type(el) is list:
            _iterate_list(
                el,
                "*",
                other_states,
                output_splitter=output_splitter,
                state_fields=state_fields,
            )
    else:
        raise PydraStateError("splitter has to be a string, a tuple or a list")
    if i > 0:
        output_splitter.append(current_sign)


def _iterate_list(element, sign, other_states, output_splitter, state_fields=True):
    """Iterate over list (used in the splitter2rpn to get recursion)."""
    for i, el in enumerate(element):
        _ordering(
            deepcopy(el),
            i,
            current_sign=sign,
            other_states=other_states,
            output_splitter=output_splitter,
            state_fields=state_fields,
        )


def converter_groups_to_input(group_for_inputs):
    """
    Return fields for each axis and number of all groups.

    Requires having axes for all the input fields.

    Parameters
    ----------
    group_for_inputs :
        specified axes (groups) for each input

    """
    input_for_axis = {}
    ngr = 0
    for inp, grs in group_for_inputs.items():
        for gr in ensure_list(grs):
            if gr in input_for_axis.keys():
                input_for_axis[gr].append(inp)
            else:
                ngr += 1
                input_for_axis[gr] = [inp]
    return input_for_axis, ngr


def remove_inp_from_splitter_rpn(splitter_rpn, inputs_to_remove):
    """
    Remove inputs due to combining.

    Mutates a splitter.

    Parameters
    ----------
    splitter_rpn :
        The splitter in reverse polish notation
    inputs_to_remove :
        input names that should be removed from the splitter

    """
    splitter_rpn_copy = splitter_rpn.copy()
    # reverting order
    splitter_rpn_copy.reverse()
    stack_inp = []
    stack_sgn = []
    from_last_sign = []
    for ii, el in enumerate(splitter_rpn_copy):
        # element is a sign
        if el == "." or el == "*":
            stack_sgn.append((ii, el))
            from_last_sign.append(0)
        # it's an input but not to remove
        elif el not in inputs_to_remove:
            if from_last_sign:
                from_last_sign[-1] += 1
            stack_inp.append((ii, el))
        # it'a an input that should be removed
        else:
            if not from_last_sign:
                pass
            elif from_last_sign[-1] <= 1:
                stack_sgn.pop()
                from_last_sign.pop()
            else:
                stack_sgn.pop(-1 * from_last_sign.pop())

    # creating the final splitter_rpn after combining
    remaining_elements = stack_sgn + stack_inp
    remaining_elements.sort(reverse=True)
    splitter_rpn_combined = [el for (i, el) in remaining_elements]
    return splitter_rpn_combined


def rpn2splitter(splitter_rpn):
    """
    Convert from splitter_rpn to splitter.

    Recurrent algorithm to perform the conversion.
    Every time combines pairs of input in one input,
    ends when the length is one.

    Parameters
    ----------
    splitter_rpn :
        splitter in reverse polish notation

    Returns
    -------
    splitter :
        splitter in the standard/original form

    """
    if splitter_rpn == []:
        return None
    if len(splitter_rpn) == 1:
        return splitter_rpn[0]

    splitter_rpn_copy = splitter_rpn.copy()
    signs = [".", "*"]
    splitter_modified = []

    while splitter_rpn_copy:
        el = splitter_rpn_copy.pop()
        # element is a sign
        if el in signs:
            if (
                splitter_rpn_copy[-1] not in signs
                and splitter_rpn_copy[-2] not in signs
            ):
                right, left = splitter_rpn_copy.pop(), splitter_rpn_copy.pop()
                if el == ".":
                    splitter_modified.append((left, right))
                elif el == "*":
                    splitter_modified.append([left, right])
            else:
                splitter_modified.append(el)
        else:
            splitter_modified.append(el)

    # reversing the list and combining more
    splitter_modified.reverse()
    return rpn2splitter(splitter_modified)


def add_name_combiner(combiner, name):
    """adding a node's name to each field from the combiner"""
    combiner_changed = []
    for comb in combiner:
        if "." not in comb:
            combiner_changed.append(f"{name}.{comb}")
        else:
            combiner_changed.append(comb)
    return combiner_changed


def add_name_splitter(
    splitter: ty.Union[str, ty.List[str], ty.Tuple[str, ...], None], name: str
) -> ty.Optional[ty.List[str]]:
    """adding a node's name to each field from the splitter"""
    if isinstance(splitter, str):
        return _add_name([splitter], name)[0]
    elif isinstance(splitter, list):
        return _add_name(list(splitter), name)
    elif isinstance(splitter, tuple):
        return tuple(_add_name(list(splitter), name))
    else:
        return None


def _add_name(mlist, name):
    """adding anem to each element from the list"""
    for i, elem in enumerate(mlist):
        if isinstance(elem, str):
            if "." in elem or elem.startswith("_"):
                pass
            else:
                mlist[i] = f"{name}.{mlist[i]}"
        elif isinstance(elem, list):
            mlist[i] = _add_name(elem, name)
        elif isinstance(elem, tuple):
            mlist[i] = list(elem)
            mlist[i] = _add_name(mlist[i], name)
            mlist[i] = tuple(mlist[i])
    return mlist


def flatten(vals, cur_depth=0, max_depth=None):
    """Flatten a list of values."""
    if max_depth is None:
        max_depth = len(list(input_shape(vals)))
    values = []
    if cur_depth >= max_depth:
        values.append([vals])
    else:
        for val in vals:
            if isinstance(val, (list, tuple)):
                values.append(flatten(val, cur_depth + 1, max_depth))
            else:
                values.append([val])
    return itertools.chain.from_iterable(values)


def iter_splits(iterable, keys):
    """Generate splits."""
    for iter in list(iterable):
        yield dict(zip(keys, list(flatten(iter, max_depth=1000))))


def input_shape(inp, container_ndim=1):
    """Get input shape, depends on the container dimension, if not specify it is assumed to be 1"""
    # TODO: have to be changed for inner splitter (sometimes different length)
    container_ndim -= 1
    shape = [len(inp)]
    last_shape = None
    for value in inp:
        if isinstance(value, list) and container_ndim > 0:
            cur_shape = input_shape(value, container_ndim)
            if last_shape is None:
                last_shape = cur_shape
            elif last_shape != cur_shape:
                last_shape = None
                break
        else:
            last_shape = None
            break
    if last_shape is not None:
        shape.extend(last_shape)
    return tuple(shape)


def splits_groups(splitter_rpn, combiner=None, inner_inputs=None):
    """splits inputs to groups (axes) and creates stacks for these groups
    This is used to specify which input can be combined.
    """
    if not splitter_rpn:
        return [], {}, [], []
    stack = []
    keys = []
    groups = {}
    group_count = None
    if not combiner:
        combiner = []
    if inner_inputs:
        previous_states_ind = {
            f"_{v.name}": v.keys_final for v in inner_inputs.values()
        }
        inner_inputs = {k: v for k, v in inner_inputs.items() if k in splitter_rpn}
    else:
        previous_states_ind = {}
        inner_inputs = {}

    # when splitter is a single element (no operators)
    if len(splitter_rpn) == 1:
        op_single = splitter_rpn[0]
        return _single_op_splits_groups(op_single, combiner, inner_inputs, groups)

    # len(splitter_rpn) > 1
    # iterating splitter_rpn
    for token in splitter_rpn:
        if token in [".", "*"]:
            terms = {}
            terms["R"] = stack.pop()
            terms["L"] = stack.pop()

            # checking if opL/R are strings
            trm_str = {"L": False, "R": False}
            oldgroups = {}

            for lr in ["L", "R"]:
                if isinstance(terms[lr], str):
                    trm_str[lr] = True
                else:
                    oldgroups[lr] = terms[lr]

            if token == ".":
                if all(trm_str.values()):
                    if group_count is None:
                        group_count = 0
                    else:
                        group_count += 1
                    oldgroup = groups[terms["L"]] = groups[terms["R"]] = group_count
                elif trm_str["R"]:
                    groups[terms["R"]] = oldgroups["L"]
                    oldgroup = oldgroups["L"]
                elif trm_str["L"]:
                    groups[terms["L"]] = oldgroups["R"]
                    oldgroup = oldgroups["R"]
                else:
                    if len(ensure_list(oldgroups["L"])) != len(
                        ensure_list(oldgroups["R"])
                    ):
                        raise ValueError(
                            "Operands do not have same shape "
                            "(left one is {}d and right one is {}d.".format(
                                len(ensure_list(oldgroups["L"])),
                                len(ensure_list(oldgroups["R"])),
                            )
                        )
                    oldgroup = oldgroups["L"]
                    # dj: changing axes for Right part of the scalar op.
                    for k, v in groups.items():
                        if v in ensure_list(oldgroups["R"]):
                            groups[k] = ensure_list(oldgroups["L"])[
                                ensure_list(oldgroups["R"]).index(v)
                            ]
            else:  # if token == "*":
                if all(trm_str.values()):
                    if group_count is None:
                        group_count = 0
                    else:
                        group_count += 1
                    groups[terms["L"]] = group_count
                    group_count += 1
                    groups[terms["R"]] = group_count
                    oldgroup = [groups[terms["L"]], groups[terms["R"]]]
                elif trm_str["R"]:
                    group_count += 1
                    groups[terms["R"]] = group_count
                    oldgroup = ensure_list(oldgroups["L"]) + [groups[terms["R"]]]
                elif trm_str["L"]:
                    group_count += 1
                    groups[terms["L"]] = group_count
                    oldgroup = [groups[terms["L"]]] + ensure_list(oldgroups["R"])
                else:
                    oldgroup = ensure_list(oldgroups["L"]) + ensure_list(oldgroups["R"])

            # creating list of keys
            if trm_str["L"]:
                if terms["L"].startswith("_"):
                    keys = previous_states_ind[terms["L"]] + keys
                else:
                    keys.insert(0, terms["L"])
            if trm_str["R"]:
                if terms["R"].startswith("_"):
                    keys += previous_states_ind[terms["R"]]
                else:
                    keys.append(terms["R"])

            pushgroup = oldgroup
            stack.append(pushgroup)

        else:  # name of one of the inputs
            stack.append(token)

    groups_stack = stack.pop()
    if isinstance(groups_stack, int):
        groups_stack = [groups_stack]
    if inner_inputs:
        groups_stack = [[], groups_stack]
    else:
        groups_stack = [groups_stack]

    if combiner:
        (
            keys_final,
            groups_final,
            groups_stack_final,
            combiner_all,
        ) = combine_final_groups(combiner, groups, groups_stack, keys)
        return keys_final, groups_final, groups_stack_final, combiner_all
    else:
        return keys, groups, groups_stack, []


def _single_op_splits_groups(op_single, combiner, inner_inputs, groups):
    """splits_groups function if splitter is a singleton"""
    if op_single in inner_inputs:
        # TODO: have to be changed if differ length
        # TODO: i think I don't want to add here from left part
        # keys = inner_inputs[op_single].keys_final + [op_single]
        keys = [op_single]
        groups[op_single], groups_stack = 0, [[], [0]]
    else:
        keys = [op_single]
        groups[op_single], groups_stack = 0, [[0]]
    if combiner:
        if combiner == [op_single]:
            return [], {}, [], combiner
        else:
            # TODO: probably not needed, should be already check by st.combiner_validation
            raise PydraStateError(
                f"all fields from the combiner have to be in splitter_rpn: {[op_single]}, "
                f"but combiner: {combiner} is set"
            )
    else:
        return keys, groups, groups_stack, []


def combine_final_groups(combiner, groups, groups_stack, keys):
    """Combine the final groups."""
    input_for_groups, _ = converter_groups_to_input(groups)
    combiner_all = []
    for comb in combiner:
        for gr in ensure_list(groups[comb]):
            combiner_all += input_for_groups[gr]
    combiner_all = list(set(combiner_all))
    combiner_all.sort()

    # groups that were removed (so not trying to remove twice)
    grs_removed = []
    groups_stack_final = deepcopy(groups_stack)
    for comb in combiner:
        grs = groups[comb]
        for gr in ensure_list(grs):
            if gr in groups_stack_final[-1]:
                grs_removed.append(gr)
                groups_stack_final[-1].remove(gr)
            elif gr in grs_removed:
                pass
            else:
                raise PydraStateError(
                    "input {} not ready to combine, you have to combine {} "
                    "first".format(comb, groups_stack[-1])
                )
    groups_final = {inp: gr for (inp, gr) in groups.items() if inp not in combiner_all}
    gr_final = set()
    for el in groups_final.values():
        gr_final.update(ensure_list(el))
    gr_final = list(gr_final)
    map_gr_nr = {nr: i for (i, nr) in enumerate(sorted(gr_final))}
    groups_final_map = {}
    for inp, gr in groups_final.items():
        if isinstance(gr, int):
            groups_final_map[inp] = map_gr_nr[gr]
        elif isinstance(gr, list):
            groups_final_map[inp] = [map_gr_nr[el] for el in gr]
        else:
            raise Exception("gr should be an int or a list, something wrong")
    for i, groups_l in enumerate(groups_stack_final):
        groups_stack_final[i] = [map_gr_nr[gr] for gr in groups_l]

    keys_final = [key for key in keys if key not in combiner_all]
    # TODO: not sure if I have to calculate and return keys, groups, groups_stack
    return keys_final, groups_final_map, groups_stack_final, combiner_all


def map_splits(split_iter, inputs, container_ndim=None):
    """generate a dictionary of inputs prescribed by the splitter."""
    if container_ndim is None:
        container_ndim = {}
    for split in split_iter:
        yield {
            k: list(
                flatten(ensure_list(inputs[k]), max_depth=container_ndim.get(k, None))
            )[v]
            for k, v in split.items()
        }


def inputs_types_to_dict(name, inputs):
    """Convert type.Inputs to dictionary."""
    # dj: any better option?
    input_names = [field for field in attrs_values(inputs) if field != "_func"]
    inputs_dict = {}
    for field in input_names:
        inputs_dict[f"{name}.{field}"] = getattr(inputs, field)
    return inputs_dict


def unwrap_splitter(
    splitter: ty.Union[str, ty.List[str], ty.Tuple[str, ...]],
) -> ty.Iterable[str]:
    """Unwraps a splitter into a flat list of fields that are split over, i.e.
    [("a", "b"), "c"] -> ["a", "b", "c"]

    Parameters
    ----------
    splitter: str or list[str] or tuple[str, ...]
        the splitter task to unwrap

    Returns
    -------
    unwrapped : ty.Iterable[str]
        the field names listed in the splitter
    """
    if isinstance(splitter, str):
        return [splitter]
    else:
        return itertools.chain(*(unwrap_splitter(s) for s in splitter))


class PydraStateError(Exception):
    """Custom error for Pydra State"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)


op = {".": zip, "*": itertools.product}



import typing as ty
import logging
from pydra.engine.job import Job
from pydra.workers import base

if ty.TYPE_CHECKING:
    from pydra.engine.result import Result


logger = logging.getLogger("pydra.worker")


class DebugWorker(base.Worker):
    """A worker to execute linearly."""

    def run(
        self,
        job: "Job[base.TaskType]",
        rerun: bool = False,
    ) -> "Result":
        """Run a job."""
        return job.run(rerun=rerun)

    def close(self):
        """Return whether the job is finished."""


# Alias so it can be referred to as debug.Worker
Worker = DebugWorker



from pydra.compose import python, workflow


@python.define
def Add(x: float, y: float) -> float:
    return x + y


@python.define
def Divide(x: float, y: float) -> float:
    return x / y


@python.define
def SafeDivide(x: float, y: float) -> float:
    if y == 0:
        return float("nan")
    return x / y


@python.define
def Subtract(x: float, y: float) -> float:
    return x - y


@workflow.define
def UnsafeDivisionWorkflow(a: float, b: float, denominator: float) -> float:
    """Adds 'a' and 'b' together, divides by 'denominator', and then subtracts 'b' from
    the output. Division by 0 is not guarded against so the workflow will fail if
    the value passed to the 'denominator' parameter is 0.

    Parameters
    ----------
    a : float
        The first number to add.
    b : float
        The second number to add.
    denominator : float
        The number to divide the sum of 'a' and 'b' by.

    Returns
    -------
    out : float
        The result of subtracting 'b' from the result of dividing the sum of 'a' and
        'b' by 'denominator'.
    """
    add = workflow.add(Add(x=a, y=b))
    divide = workflow.add(Divide(x=add.out, y=denominator))
    subtract = workflow.add(Subtract(x=divide.out, y=b))
    return subtract.out


@workflow.define
def SafeDivisionWorkflow(a: float, b: float, denominator: float) -> float:
    """Adds 'a' and 'b' together, divides by 'denominator', and then subtracts 'b' from
    the output. Division by 0 is not guarded against so the workflow will fail if
    the value passed to the 'denominator' parameter is 0.

    Parameters
    ----------
    a : float
        The first number to add.
    b : float
        The second number to add.
    denominator : float
        The number to divide the sum of 'a' and 'b' by.

    Returns
    -------
    out : float
        The result of subtracting 'b' from the result of dividing the sum of 'a' and
        'b' by 'denominator'.
    """
    add = workflow.add(Add(x=a, y=b))
    divide = workflow.add(SafeDivide(x=add.out, y=denominator))
    subtract = workflow.add(Subtract(x=divide.out, y=b))
    return subtract.out


@python.define
def TenToThePower(p: int) -> int:
    return 10**p



import asyncio
import sys
import json
import re
import attrs
import typing as ty
from tempfile import gettempdir
from pathlib import Path
from shutil import copyfile, which
import random
import logging
from pydra.engine.job import Job, save, load_job
from pydra.workers import base

logger = logging.getLogger("pydra.worker")

if ty.TYPE_CHECKING:
    from pydra.engine.result import Result


@attrs.define
class SgeWorker(base.Worker):
    """A worker to execute tasks on SLURM systems. Initialize SGE Worker.

    Parameters
    ----------
    poll_delay : seconds
        Delay between polls to slurmd
    qsub_args : str
        Additional qsub arguments
    max_jobs : int
        Maximum number of submitted jobs
    write_output_files : bool
        Turns on/off writing to output files for individual tasks
    max_job_array_length : int
        Number of jobs an SGE job array can hold
    indirect_submit_host : str
        Name of a submit node in the SGE cluster through which to run SGE qsub commands
    max_threads : int
        Maximum number of threads that will be scheduled for SGE submission at once
    poll_for_result_file : bool
        If true, a job is complete when its _result.pklz file exists
        If false, a job is complete when its job array is indicated complete by qstat/qacct polling
    default_threads_per_task : int
        Sets the number of slots SGE should request for a job if sgeThreads
        is not a field in the job input_spec
    polls_before_checking_evicted : int
        Number of poll_delays before running qacct to check if a job has been evicted by SGE
    collect_jobs_delay : int
        Number of seconds to wait for the list of jobs for a job array to fill
    """

    _cmd = "qsub"
    _sacct_re = re.compile(
        "(?P<jobid>\\d*) +(?P<status>\\w*)\\+? +" "(?P<exit_code>\\d+):\\d+"
    )

    poll_delay: int = attrs.field(default=1, converter=base.ensure_non_negative)
    qsub_args: str = ""
    write_output_files: bool = True
    max_job_array_length: int = 50
    indirect_submit_host: str | None = None
    max_threads: int | None = None
    poll_for_result_file: bool = True
    default_threads_per_task: int = 1
    polls_before_checking_evicted: int = 60
    collect_jobs_delay: int = 30
    default_qsub_args: str = ""
    max_mem_free: int | None = None
    error: dict[str, ty.Any] = attrs.field(factory=dict, init=False)
    tasks_to_run_by_threads_requested: dict[str, ty.Any] = attrs.field(
        factory=dict, init=False
    )
    output_by_jobid: dict[str, ty.Any] = attrs.field(factory=dict, init=False)
    jobid_by_task_uid: dict[str, ty.Any] = attrs.field(factory=dict, init=False)
    threads_used: dict[str, ty.Any] = attrs.field(factory=dict, init=False)
    job_completed_by_jobid: dict[str, ty.Any] = attrs.field(factory=dict, init=False)
    result_files_by_jobid: dict[str, ty.Any] = attrs.field(factory=dict, init=False)
    job_pkls_rerun: dict[str, ty.Any] = attrs.field(factory=dict, init=False)

    _INTERNAL_DICT_ATTRS = [
        "error",
        "tasks_to_run_by_threads_requested",
        "output_by_jobid",
        "jobid_by_task_uid",
        "threads_used",
        "job_completed_by_jobid",
        "result_files_by_jobid",
        "job_pkls_rerun",
    ]

    def __getstate__(self) -> dict[str, ty.Any]:
        """Return state for pickling."""
        state = super().__getstate__()
        for atr in self._INTERNAL_DICT_ATTRS:
            del state[atr]
        return state

    def __setstate__(self, state: dict[str, ty.Any]):
        """Set state for unpickling."""
        super().__setstate__(state)
        for atr in self._INTERNAL_DICT_ATTRS:
            setattr(self, atr, {})

    def _prepare_runscripts(self, job, interpreter="/bin/sh", rerun=False):
        if isinstance(job, Job):
            cache_root = job.cache_root
            ind = None
            uid = job.uid
            try:
                task_qsub_args = job.qsub_args
            except Exception:
                task_qsub_args = self.default_qsub_args
        else:
            ind = job[0]
            cache_root = job[-1].cache_root
            uid = f"{job[-1].uid}_{ind}"
            try:
                task_qsub_args = job[-1].qsub_args
            except Exception:
                task_qsub_args = self.default_qsub_args

        script_dir = cache_root / f"{self.plugin_name()}_scripts" / uid
        script_dir.mkdir(parents=True, exist_ok=True)
        if ind is None:
            if not (script_dir / "_job.pklz").exists():
                save(script_dir, job=job)
        else:
            copyfile(job[1], script_dir / "_job.pklz")

        job_pkl = script_dir / "_job.pklz"
        if not job_pkl.exists() or not job_pkl.stat().st_size:
            raise Exception("Missing or empty job!")

        batchscript = script_dir / f"batchscript_{uid}.job"

        if task_qsub_args not in self.tasks_to_run_by_threads_requested:
            self.tasks_to_run_by_threads_requested[task_qsub_args] = []
        self.tasks_to_run_by_threads_requested[task_qsub_args].append(
            (str(job_pkl), ind, rerun)
        )

        return (
            script_dir,
            batchscript,
            job_pkl,
            ind,
            job.cache_dir,
            task_qsub_args,
        )

    async def get_tasks_to_run(self, task_qsub_args, mem_free):
        # Extract the first N tasks to run
        if mem_free is not None and self.max_mem_free is not None:
            max_job_array_length = min(
                self.max_job_array_length, int(self.max_mem_free / mem_free)
            )
        else:
            max_job_array_length = self.max_job_array_length
        tasks_to_run_copy, self.tasks_to_run_by_threads_requested[task_qsub_args] = (
            self.tasks_to_run_by_threads_requested[task_qsub_args][
                :max_job_array_length
            ],
            self.tasks_to_run_by_threads_requested[task_qsub_args][
                max_job_array_length:
            ],
        )
        return tasks_to_run_copy

    async def check_for_results_files(self, jobid, threads_requested):
        for job in list(self.result_files_by_jobid[jobid]):
            if self.result_files_by_jobid[jobid][job].exists():
                del self.result_files_by_jobid[jobid][job]
                self.threads_used -= threads_requested

    async def run(self, job: Job[base.TaskType], rerun: bool = False) -> "Result":
        """Worker submission API."""
        (
            script_dir,
            batch_script,
            job_pkl,
            ind,
            cache_dir,
            task_qsub_args,
        ) = self._prepare_runscripts(job, rerun=rerun)
        if (script_dir / script_dir.parts[1]) == gettempdir():
            logger.warning("Temporary directories may not be shared across computers")
        interpreter = "/bin/sh"
        threads_requested = self.default_threads_per_task
        if "smp" in task_qsub_args:
            smp_index = task_qsub_args.split().index("smp")
            if (
                smp_index + 1 < len(task_qsub_args.split())
                and task_qsub_args.split()[smp_index + 1].isdigit()
            ):
                threads_requested = int(task_qsub_args.split()[smp_index + 1])
        # Get the amount of mem_free requested for the job
        mem_free = None
        if "mem_free" in task_qsub_args:
            mem_free_cmd = [
                word for word in task_qsub_args.split() if word.startswith("mem_free")
            ][0]
            if len(re.findall(r"\d+", mem_free_cmd)) > 0:
                mem_free = int(re.findall(r"\d+", mem_free_cmd)[0])

        if (
            len(self.tasks_to_run_by_threads_requested.get(task_qsub_args))
            <= self.max_job_array_length
        ):
            await asyncio.sleep(self.collect_jobs_delay)
        tasks_to_run = await self.get_tasks_to_run(task_qsub_args, mem_free)

        if mem_free is not None:
            summed_mem_free_cmd = re.sub(
                str(mem_free), str(len(tasks_to_run) * mem_free), mem_free_cmd
            )
            task_qsub_args = re.sub(mem_free_cmd, summed_mem_free_cmd, task_qsub_args)

        if len(tasks_to_run) > 0:
            if self.max_threads is not None:
                while self.threads_used > self.max_threads - threads_requested * len(
                    tasks_to_run
                ):
                    await asyncio.sleep(self.poll_delay)
            self.threads_used += threads_requested * len(tasks_to_run)

            python_string = f"""import sys; from pydra.engine.job import load_and_run; \
                job_pkls={[task_tuple for task_tuple in tasks_to_run]}; \
                task_index=int(sys.argv[1])-1; \
                load_and_run(job_pkls[task_index][0], rerun=job_pkls[task_index][1])"""
            bcmd_job = "\n".join(
                (
                    f"#!{interpreter}",
                    f"{sys.executable} {Path(batch_script).with_suffix('.py')}"
                    + " $SGE_TASK_ID",
                )
            )

            bcmd_py = python_string

            # Better runtime when the python contents are written to file
            # rather than given by cmdline arg -c
            with Path(batch_script).with_suffix(".py").open("wt") as fp:
                fp.write(bcmd_py)

            with batch_script.open("wt") as fp:
                fp.writelines(bcmd_job)

            script_dir = job.cache_root / f"{self.plugin_name()}_scripts" / job.uid
            script_dir.mkdir(parents=True, exist_ok=True)
            sargs = ["-t"]
            sargs.append(f"1-{len(tasks_to_run)}")
            sargs = sargs + task_qsub_args.split()

            jobname = re.search(r"(?<=-N )\S+", task_qsub_args)

            if not jobname:
                jobname = ".".join((job.name, job.uid))
                sargs.append("-N")
                sargs.append(jobname)
            output = re.search(r"(?<=-o )\S+", self.qsub_args)

            if not output:
                output_file = str(script_dir / "sge-%j.out")
                if self.write_output_files:
                    sargs.append("-o")
                    sargs.append(output_file)
            error = re.search(r"(?<=-e )\S+", self.qsub_args)
            if not error:
                error_file = str(script_dir / "sge-%j.out")
                if self.write_output_files:
                    sargs.append("-e")
                    sargs.append(error_file)
            else:
                error_file = None
            sargs.append(str(batch_script))

            await asyncio.sleep(random.uniform(0, 5))

            jobid = await self.submit_array_job(sargs, tasks_to_run, error_file)

            if self.poll_for_result_file:
                self.result_files_by_jobid[jobid] = {}
                for job_pkl, ind, rerun in tasks_to_run:
                    job = load_job(job_pkl=job_pkl, ind=ind)
                    self.result_files_by_jobid[jobid][job] = (
                        job.cache_dir / "_result.pklz"
                    )

            poll_counter = 0
            while True:
                # 3 possibilities
                # False: job is still pending/working
                # True: job is complete
                # Exception: Polling / job failure
                # done = await self._poll_job(jobid)
                if self.poll_for_result_file:
                    if len(self.result_files_by_jobid[jobid]) > 0:
                        for job in list(self.result_files_by_jobid[jobid]):
                            if self.result_files_by_jobid[jobid][job].exists():
                                del self.result_files_by_jobid[jobid][job]
                                self.threads_used -= threads_requested

                    else:
                        exit_status = await self._verify_exit_code(jobid)
                        if exit_status == "ERRORED":
                            jobid = await self._rerun_job_array(
                                job.cache_root,
                                job.uid,
                                sargs,
                                tasks_to_run,
                                error_file,
                                jobid,
                            )
                        else:
                            for job_pkl, ind, rerun in tasks_to_run:
                                if job_pkl in self.job_pkls_rerun:
                                    del self.job_pkls_rerun[job_pkl]
                            return True

                    if poll_counter >= self.polls_before_checking_evicted:
                        # Checking for evicted for jobid
                        exit_status = await self._verify_exit_code(jobid)
                        if exit_status == "ERRORED":
                            jobid = await self._rerun_job_array(
                                job.cache_root,
                                job.uid,
                                sargs,
                                tasks_to_run,
                                error_file,
                                jobid,
                            )
                        poll_counter = 0
                    poll_counter += 1
                    await asyncio.sleep(self.poll_delay)
                else:
                    done = await self._poll_job(jobid, job.cache_root)
                    if done:
                        if done == "ERRORED":  # If the SGE job was evicted, rerun it
                            jobid = await self._rerun_job_array(
                                job.cache_root,
                                job.uid,
                                sargs,
                                tasks_to_run,
                                error_file,
                                jobid,
                            )
                        else:
                            self.job_completed_by_jobid[jobid] = True
                            self.threads_used -= threads_requested * len(tasks_to_run)
                            return True
                    # Don't poll exactly on the same interval to avoid overloading SGE
                    await asyncio.sleep(
                        random.uniform(max(0, self.poll_delay - 2), self.poll_delay + 2)
                    )

    async def _rerun_job_array(
        self, cache_root, uid, sargs, tasks_to_run, error_file, evicted_jobid
    ):
        for job_pkl, ind, rerun in tasks_to_run:
            sge_task = load_job(job_pkl=job_pkl, ind=ind)
            application_job_pkl = sge_task.cache_dir / "_job.pklz"
            if (
                not application_job_pkl.exists()
                or load_job(job_pkl=application_job_pkl).result() is None
                or load_job(job_pkl=application_job_pkl).result().errored
            ):
                self.job_pkls_rerun[job_pkl] = None
                info_file = cache_root / f"{sge_task.uid}_info.json"
                if info_file.exists():
                    checksum = json.loads(info_file.read_text())["checksum"]
                    if (cache_root / f"{checksum}.lock").exists():
                        # for pyt3.8 we could use missing_ok=True
                        (cache_root / f"{checksum}.lock").unlink()
                    # Maybe wait a little to check if _error.pklz exists - not getting found immediately

        # If the previous job array failed, run the array's script again and get the new jobid
        jobid = await self.submit_array_job(sargs, tasks_to_run, error_file)
        self.result_files_by_jobid[jobid] = self.result_files_by_jobid[evicted_jobid]
        return jobid

    async def submit_array_job(self, sargs, tasks_to_run, error_file):
        if self.indirect_submit_host is not None:
            indirect_submit_host_prefix = []
            indirect_submit_host_prefix.append("ssh")
            indirect_submit_host_prefix.append(self.indirect_submit_host)
            indirect_submit_host_prefix.append('""export SGE_ROOT=/opt/sge;')
            rc, stdout, stderr = await base.read_and_display_async(
                *indirect_submit_host_prefix,
                str(Path(which("qsub")).parent / "qsub"),
                *sargs,
                '""',
                hide_display=True,
            )
        else:
            rc, stdout, stderr = await base.read_and_display_async(
                "qsub", *sargs, hide_display=True
            )
        jobid = re.search(r"\d+", stdout)
        if rc:
            raise RuntimeError(f"Error returned from qsub: {stderr}")
        elif not jobid:
            raise RuntimeError("Could not extract job ID")
        jobid = jobid.group()
        self.output_by_jobid[jobid] = (rc, stdout, stderr)

        for job_pkl, ind, rerun in tasks_to_run:
            self.jobid_by_task_uid[Path(job_pkl).parent.name] = jobid

        if error_file:
            error_file = str(error_file).replace("%j", jobid)
        self.error[jobid] = str(error_file).replace("%j", jobid)
        return jobid

    async def get_output_by_job_pkl(self, job_pkl):
        jobid = self.jobid_by_task_uid.get(job_pkl.parent.name)
        while jobid is None:
            jobid = self.jobid_by_task_uid.get(job_pkl.parent.name)
            await asyncio.sleep(1)
        job_output = self.output_by_jobid.get(jobid)
        while job_output is None:
            job_output = self.output_by_jobid.get(jobid)
            await asyncio.sleep(1)
        return job_output

    async def _submit_job(
        self,
        batchscript,
        name,
        uid,
        cache_root,
        job_pkl,
        ind,
        cache_dir,
        task_qsub_args,
    ):
        """Coroutine that submits job runscript and polls job until completion or error."""
        await self._submit_jobs(
            batchscript,
            name,
            uid,
            cache_root,
            cache_dir,
            task_qsub_args,
        )
        if self.poll_for_result_file:
            while True:
                result_file = cache_dir / "_result.pklz"
                if result_file.exists() and str(job_pkl) not in self.job_pkls_rerun:
                    return True
                await asyncio.sleep(self.poll_delay)
        else:
            rc, stdout, stderr = await self.get_output_by_job_pkl(job_pkl)
            while True:
                jobid = self.jobid_by_task_uid.get(job_pkl.parent.name)
                if self.job_completed_by_jobid.get(jobid):
                    return True
                else:
                    await asyncio.sleep(self.poll_delay)

    async def _poll_job(self, jobid, cache_root):
        cmd = ("qstat", "-j", jobid)
        logger.debug(f"Polling job {jobid}")
        rc, stdout, stderr = await base.read_and_display_async(*cmd, hide_display=True)

        if not stdout:
            # job is no longer running - check exit code
            status = await self._verify_exit_code(jobid)
            return status
        return False

    async def _verify_exit_code(self, jobid):
        cmd = ("qacct", "-j", jobid)
        rc, stdout, stderr = await base.read_and_display_async(*cmd, hide_display=True)
        if not stdout:
            await asyncio.sleep(10)
            rc, stdout, stderr = await base.read_and_display_async(
                *cmd, hide_display=True
            )

        # job is still pending/working
        if re.match(r"error: job id .* not found", stderr):
            return False

        if not stdout:
            return "ERRORED"

        # Read the qacct stdout into dictionary stdout_dict
        for line in stdout.splitlines():
            line_split = line.split()
            if len(line_split) > 1:
                if line_split[0] == "failed":
                    if not line_split[1].isdigit():
                        return "ERRORED"
                    elif not int(line_split[1]) == 0:
                        return "ERRORED"
        return True


# Alias so it can be referred to as sge.Worker
Worker = SgeWorker



from __future__ import annotations
import typing as ty
import attrs
from pydra.compose.base import (
    Arg,
    Out,
    NO_DEFAULT,
)
from pydra.utils.typing import is_optional
from pydra.utils.general import wrap_text


@attrs.define(kw_only=True)
class arg(Arg):
    """An input field that specifies a command line argument

    Parameters
    ----------
    help: str
        A short description of the input field.
    default : Any, optional
        the default value for the argument
    mandatory: bool, optional
        If True user has to provide a value for the field, by default it is False
    allowed_values: list, optional
        List of allowed values for the field.
    requires: list, optional
        List of field names that are required together with the field.
    copy_mode: File.CopyMode, optional
        The mode of copying the file, by default it is File.CopyMode.any
    copy_collation: File.CopyCollation, optional
        The collation of the file, by default it is File.CopyCollation.any
    copy_ext_decomp: File.ExtensionDecomposition, optional
        The extension decomposition of the file, by default it is
        File.ExtensionDecomposition.single
    readonly: bool, optional
        If True the input field can’t be provided by the user but it aggregates other
        input fields (for example the fields with argstr: -o {fldA} {fldB}), by default
        it is False
    type: type, optional
        The type of the field, by default it is Any
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
        from name to field, by default it is None
    argstr: str, optional
        A flag or string that is used in the command before the value, e.g. -v or
        -v {inp_field}, but it could be and empty string, “”, in which case the value is
        just printed to the command line. If … are used, e.g. -v…,
        the flag is used before every element if a list is provided as a value. If the
        argstr is None, the field is not part of the command.
    position: int, optional
        Position of the field in the command, could be nonnegative or negative integer.
        If nothing is provided the field will be inserted between all fields with
        nonnegative positions and fields with negative positions.
    sep: str, optional
        A separator if a sequence type is provided as a value, by default " ".
    container_path: bool, optional
        If True a path will be consider as a path inside the container (and not as a
        local path, by default it is False
    formatter: function, optional
        If provided the argstr of the field is created using the function. This function
        can for example be used to combine several inputs into one command argument. The
        function can take field (this input field will be passed to the function),
        inputs (entire inputs will be passed) or any input field name (a specific input
        field will be sent).
    """

    argstr: str | None = ""
    position: int | None = None
    sep: str = " "
    allowed_values: list | None = None
    container_path: bool = False  # IS THIS STILL USED??
    formatter: ty.Callable | None = None

    def _additional_descriptors(self, as_input: bool = False, **kwargs) -> str:
        if not self.argstr or not as_input:
            return super()._additional_descriptors(as_input=as_input, **kwargs)
        descriptors = [f"{self.argstr!r}"]
        descriptors.extend(super()._additional_descriptors(as_input=as_input, **kwargs))
        return descriptors

    def __lt__(self, other: "arg") -> bool:
        """Compare two fields based on their position"""
        if self.position is None and other.position is None:
            return super().__lt__(other)
        elif self.position is None:
            return False
        elif other.position is None:
            return True
        else:
            assert self.position != other.position, "positions should be unique"
            if self.position < 0 and other.position < 0:
                return self.position > other.position
            elif self.position < 0:
                return False
            elif other.position < 0:
                return True
            return self.position < other.position


@attrs.define(kw_only=True)
class out(Out):
    """An output field that specifies a command line argument

    Parameters
    ----------
    callable : Callable, optional
        If provided the output file name (or list of file names) is created using the
        function. The function can take field (the specific output field will be passed
        to the function), cache_dir (task cache_dir will be used), stdout, stderr
        (stdout and stderr of the task will be sent) inputs (entire inputs will be
        passed) or any input field name (a specific input field will be sent).
    """

    callable: ty.Callable | None = attrs.field(default=None)

    def __attrs_post_init__(self):
        # Set type from return annotation of callable if not set
        if self.type is ty.Any and self.callable:
            self.type = ty.get_type_hints(self.callable).get("return", ty.Any)

    @callable.validator
    def _callable_validator(self, _, value):

        if value:
            if not callable(value):
                raise ValueError(f"callable must be a function, not {value!r}")
        elif (
            self.mandatory
            and not getattr(self, "path_template", None)
            and self.name
            not in [
                "return_code",
                "stdout",
                "stderr",
            ]
        ):  # shell.Outputs.BASE_ATTRS
            raise ValueError(
                "A shell output field must have either a callable or a path_template"
            )


@attrs.define(kw_only=True)
class outarg(arg, Out):
    """An input field that specifies where to save the output file

    Parameters
    ----------
    help: str
        A short description of the input field.
    default : Any, optional
        the default value for the argument
    mandatory: bool, optional
        If True user has to provide a value for the field, by default it is False
    allowed_values: list, optional
        List of allowed values for the field.
    requires: list, optional
        List of field names that are required together with the field.
    copy_mode: File.CopyMode, optional
        The mode of copying the file, by default it is File.CopyMode.any
    copy_collation: File.CopyCollation, optional
        The collation of the file, by default it is File.CopyCollation.any
    copy_ext_decomp: File.ExtensionDecomposition, optional
        The extension decomposition of the file, by default it is
        File.ExtensionDecomposition.single
    readonly: bool, optional
        If True the input field can’t be provided by the user but it aggregates other
        input fields (for example the fields with argstr: -o {fldA} {fldB}), by default
        it is False
    type: type, optional
        The type of the field, by default it is Any
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
        from name to field, by default it is None
    argstr: str, optional
        A flag or string that is used in the command before the value, e.g. -v or
        -v {inp_field}, but it could be and empty string, “”. If … are used, e.g. -v…,
        the flag is used before every element if a list is provided as a value. If no
        argstr is used the field is not part of the command.
    position: int, optional
        Position of the field in the command line, could be nonnegative or negative integer.
        If nothing is provided the field will be inserted between all fields with
        nonnegative positions and fields with negative positions.
    sep: str, optional
        A separator if a list is provided as a value.
    container_path: bool, optional
        If True a path will be consider as a path inside the container (and not as a
        local path, by default it is False
    formatter: function, optional
        If provided the argstr of the field is created using the function. This function
        can for example be used to combine several inputs into one command argument. The
        function can take field (this input field will be passed to the function),
        inputs (entire inputs will be passed) or any input field name (a specific input
        field will be sent).
    path_template: str, optional
        The template used to specify where the output file will be written to can use
        other fields, e.g. {file1}. Used in order to create an output definition.
    """

    path_template: str | None = attrs.field(default=None)
    keep_extension: bool = attrs.field(default=True)

    @path_template.validator
    def _validate_path_template(self, attribute, value):
        if value:
            if self.default not in (NO_DEFAULT, True, None):
                raise ValueError(
                    f"path_template ({value!r}) can only be provided when there is no "
                    f"default value provided ({self.default!r})"
                )

    def markdown_listing(
        self,
        line_width: int = 79,
        help_indent: int = 4,
        as_input: bool = False,
        **kwargs,
    ):
        """Get the listing for the field in markdown-like format

        Parameters
        ----------
        line_width: int
            The maximum line width for the output, by default it is 79
        help_indent: int
            The indentation for the help text, by default it is 4
        as_input: bool
            Whether to format the field as an input or output if it can be both, by default
            it is False
        **kwargs: Any
            Additional arguments to allow it to be duck-typed with extension classes

        Returns
        -------
        str
            The listing for the field in markdown-like format
        """
        if not as_input:
            return super().markdown_listing(
                width=line_width, help_indent=help_indent, **kwargs
            )

        type_str = "Path | bool"
        if is_optional(self.type):
            type_str += " | None"
            default = "None"
            help_text = wrap_text(
                self.OPTIONAL_PATH_TEMPLATE_HELP,
                width=line_width,
                indent_size=help_indent,
            )
        else:
            default = True
            help_text = wrap_text(
                self.PATH_TEMPLATE_HELP, width=line_width, indent_size=help_indent
            )
        s = f"- {self.name}: {type_str}; default = {default}"
        if self._additional_descriptors(as_input=as_input):
            s += f" ({', '.join(self._additional_descriptors(as_input=as_input))})"
        s += "\n" + help_text
        return s

    PATH_TEMPLATE_HELP = (
        "The path specified for the output file, if True, the default "
        "'path template' will be used."
    )
    OPTIONAL_PATH_TEMPLATE_HELP = PATH_TEMPLATE_HELP + (
        "If False or None, the output file will not be saved."
    )



import cloudpickle as cp
import sys
from pathlib import Path
from pydra.engine.job import load_and_run

# To avoid issues when running pytest, where the namespace package "pydra" is dropped in
# the pickling process due to it being run from inside the source tree
sys.path.append(str(Path(__file__).parent.parent))


def run_pickled(*file_paths, rerun=False):
    loaded_objects = []

    for file_path in file_paths:
        with open(file_path, "rb") as file:
            loaded_objects.append(cp.load(file))

    if len(loaded_objects) == 1:
        result = loaded_objects[0](rerun=rerun)
    elif len(loaded_objects) == 2:
        result = load_and_run(loaded_objects[0], loaded_objects[1], rerun=rerun)
    else:
        raise ValueError("Unsupported number of loaded objects")

    print(f"Result: {result}")


if __name__ == "__main__":
    rerun = False  # Default value for rerun
    file_paths = sys.argv[1:]

    if "--rerun" in file_paths:
        rerun = True
        file_paths.remove("--rerun")

    run_pickled(*file_paths, rerun=rerun)



import logging
from ._version import __version__

# import __main__

logger = logging.getLogger("pydra")


def check_latest_version():
    import etelemetry

    return etelemetry.check_available_version("nipype/pydra", __version__, lgr=logger)


# Run telemetry on import for interactive sessions, such as IPython, Jupyter notebooks, Python REPL
# if not hasattr(__main__, "__file__"):
#     from pydra.compose.base import Task

#     if Task._etelemetry_version_data is None:
#         Task._etelemetry_version_data = check_latest_version()



from pathlib import Path
import pdb
import sys

import click
import cloudpickle as cp

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
ExistingFilePath = click.Path(exists=True, dir_okay=False, resolve_path=True)


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    pass


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument("crashfile", type=ExistingFilePath)
@click.option(
    "-r", "--rerun", is_flag=True, flag_value=True, help="Rerun crashed code."
)
@click.option(
    "-d",
    "--debugger",
    type=click.Choice([None, "ipython", "pdb"]),
    help="Debugger to use when rerunning",
)
def crash(crashfile, rerun, debugger=None):
    """Display a crash file and rerun if required."""
    if crashfile.endswith(("pkl", "pklz")):
        with open(crashfile, "rb") as f:
            crash_content = cp.load(f)
        print("".join(crash_content["error message"]))

        if rerun:
            jobfile = Path(crashfile).parent / "_job.pklz"
            if jobfile.exists():
                with open(jobfile, "rb") as f:
                    job_obj = cp.load(f)

                if debugger == "ipython":
                    try:
                        from IPython.core import ultratb

                        sys.excepthook = ultratb.FormattedTB(
                            mode="Verbose", theme_name="Linux", call_pdb=True
                        )
                    except ImportError:
                        raise ImportError(
                            "'Ipython' needs to be installed to use the 'ipython' debugger"
                        )

                try:
                    job_obj.run(rerun=True)
                except Exception:  # noqa: E722
                    if debugger == "pdb":
                        pdb.post_mortem()
                    elif debugger == "ipython":
                        raise
            else:
                raise FileNotFoundError(f"Job file {jobfile} not found")
    else:
        raise ValueError("Only pickled crashfiles are supported")



"""Handle execution backends."""

import asyncio
import typing as ty
import re
import os
from pathlib import Path
from traceback import format_exc
from tempfile import mkdtemp
from copy import copy, deepcopy
from datetime import datetime
from collections import defaultdict
import attrs
import logging
from pydra.engine.graph import DiGraph
from pydra.utils.general import (
    get_fields,
    attrs_values,
)
from pydra.utils.hash import PersistentCache
from pydra.engine.lazy import LazyField
from pydra.engine.audit import Audit
from pydra.engine.job import Job
from pydra.utils.messenger import AuditFlag, Messenger
from pydra.utils.general import default_run_cache_root
from pydra.compose import workflow
from pydra.engine.state import State
from pydra.workers.base import Worker
from pydra.compose.base import Task, Outputs

logger = logging.getLogger("pydra.submitter")

if ty.TYPE_CHECKING:
    from pydra.engine.node import Node
    from pydra.engine.result import Result
    from pydra.engine.hooks import TaskHooks
    from pydra.engine.workflow import Workflow
    from pydra.environments.base import Environment


TaskType = ty.TypeVar("TaskType", bound="Task")
OutputType = ty.TypeVar("OutputType", bound="Outputs")

# Used to flag development mode of Audit
develop = False

WORKER_KWARG_FAIL_NOTE = "Attempting to instantiate worker submitter"


class Submitter:
    """Send a job to the execution backend.

    Parameters
    ----------
    cache_root : os.PathLike, optional
        Cache directory where the working directory/results for the job will be
        stored, by default None
    worker : str or Worker, optional
        The worker to use, by default "cf"
    environment: Environment, optional
        The execution environment to use, by default None
    readonly_caches : list[os.PathLike], optional
        Alternate cache locations to check for pre-computed results, by default None
    max_concurrent: int | float, optional
        Maximum number of concurrent tasks to run, by default float("inf") (unlimited)
    audit_flags : AuditFlag, optional
        Configure provenance tracking. available flags: :class:`~pydra.utils.messenger.AuditFlag`
        Default is no provenance tracking.
    messenger : :class:`Messenger` or :obj:`list` of :class:`Messenger` or None
        Messenger(s) used by Audit. Saved in the `audit` attribute.
        See available flags at :class:`~pydra.utils.messenger.Messenger`.
    messengers_args : dict[str, Any], optional
        Argument(s) used by `messegner`. Saved in the `audit` attribu
    clean_stale_locks : bool, optional
        Whether to clean stale lock files, i.e. lock files that were created before the
        start of the current run. Don't set if using a global cache where there are
        potentially multiple workflows that are running concurrently. By default (None),
        lock files will be cleaned if the *debug* worker is used
    **kwargs : dict
        Keyword arguments to pass on to the worker initialisation
    """

    cache_root: os.PathLike
    worker: Worker
    environment: "Environment | None"
    readonly_caches: list[os.PathLike]
    audit_flags: AuditFlag
    messengers: ty.Iterable[Messenger]
    messenger_args: dict[str, ty.Any]
    max_concurrent: int | float
    clean_stale_locks: bool
    run_start_time: datetime | None
    propagate_rerun: bool

    def __init__(
        self,
        /,
        cache_root: os.PathLike | None = None,
        worker: str | ty.Type[Worker] | Worker | None = "debug",
        environment: "Environment | None" = None,
        readonly_caches: list[os.PathLike] | None = None,
        audit_flags: AuditFlag = AuditFlag.NONE,
        messengers: ty.Iterable[Messenger] | None = None,
        messenger_args: dict[str, ty.Any] | None = None,
        max_concurrent: int | float = float("inf"),
        propagate_rerun: bool = True,
        clean_stale_locks: bool | None = None,
        **kwargs,
    ):

        from pydra.environments import native

        if worker is None:
            worker = "debug"

        from pydra.utils.etelemetry import check_latest_version

        if Job._etelemetry_version_data is None:
            Job._etelemetry_version_data = check_latest_version()

        self.audit = Audit(
            audit_flags=audit_flags,
            messengers=messengers,
            messenger_args=messenger_args,
            develop=develop,
        )
        if cache_root is None:
            cache_root = default_run_cache_root
        cache_root = Path(cache_root).resolve()
        cache_root.mkdir(parents=True, exist_ok=True)

        self.cache_root = cache_root
        self.readonly_caches = readonly_caches
        self.propagate_rerun = propagate_rerun
        if max_concurrent < 1 or (
            isinstance(max_concurrent, float) and max_concurrent != float("inf")
        ):
            raise ValueError(
                "'max_concurrent' arg must be a positive integer or float('inf'), "
                f"not {max_concurrent}"
            )
        self.max_concurrent = max_concurrent
        self.environment = (
            environment if environment is not None else native.Environment()
        )
        self.loop = get_open_loop()
        self._own_loop = not self.loop.is_running()
        if not isinstance(worker, Worker):
            if isinstance(worker, str):
                worker_cls = Worker.plugin(worker)
            elif issubclass(worker, Worker):
                worker_cls = worker
            else:
                raise TypeError(
                    "Worker must be a Worker object, name of a worker or a Worker "
                    f"class, not {worker}"
                )
            try:
                worker = worker_cls(**kwargs)
            except TypeError as e:
                e.add_note(WORKER_KWARG_FAIL_NOTE)
                raise
        self.worker = worker
        self.run_start_time = None
        self.clean_stale_locks = (
            clean_stale_locks
            if clean_stale_locks is not None
            else (self.worker.plugin_name() == "debug")
        )
        self.worker_kwargs = kwargs
        self.worker.loop = self.loop

    def __call__(
        self,
        task: "Task[OutputType]",
        hooks: "TaskHooks | None" = None,
        raise_errors: bool | None = None,
        rerun: bool = False,
    ) -> "Result[OutputType]":
        """Submitter run function.

        Parameters
        ----------
        task : :obj:`~pydra.compose.base.Task`
            The task to run
        hooks : :obj:`~pydra.engine.hooks.TaskHooks`, optional
            Job hooks, callable functions called as the job is setup and torn down,
            by default no functions are called at the hooks
        raise_errors : bool, optional
            Whether to raise errors, by default True if the 'debug' worker is used,
            otherwise False
        rerun : bool, optional
            Whether to force the re-computation of the job results even if existing
            results are found, by default False
        propagate_rerun : bool, optional
            Whether to propagate the rerun flag to all tasks in the workflow, by default True

        Returns
        -------
        result : Any
            The result of the job
        """
        from pydra.environments.base import Environment

        if raise_errors is None:
            raise_errors = self.worker.plugin_name() == "debug"
        if not isinstance(raise_errors, bool):
            raise TypeError(
                f"'raise_errors' must be a boolean or None, not {type(raise_errors)}"
            )

        task._check_rules()
        # If the outer job is split, create an implicit workflow to hold the split nodes
        if task._splitter:

            state = State(
                name="outer_split",
                splitter=deepcopy(task._splitter),
                combiner=deepcopy(task._combiner),
                container_ndim=deepcopy(task._container_ndim),
            )

            def wrap_type(tp):
                tp = state.nest_output_type(tp)
                tp = state.combine_state_arrays(tp)
                return tp

            output_types = {o.name: wrap_type(o.type) for o in get_fields(task.Outputs)}

            @workflow.define(outputs=output_types)
            def Split(defn: Task, output_types: dict, environment: Environment | None):
                node = workflow.add(defn, environment=environment, hooks=hooks)
                return tuple(getattr(node, o) for o in output_types)

            task = Split(
                defn=task, output_types=output_types, environment=self.environment
            )

            environment = None
        elif task._combiner:
            raise ValueError(
                f"Job {self} is marked for combining, but not splitting. "
                "Use the `split` method to split the job before combining."
            )
        else:
            environment = self.environment

        job = Job(
            task,
            submitter=self,
            name="main",
            environment=environment,
            hooks=hooks,
        )
        try:
            self.run_start_time = datetime.now()
            self.submit(job, rerun=rerun)
        except Exception as exc:
            error_msg = (
                f"Full crash report for {type(task).__name__!r} job is here: "
                + str(job.cache_dir / "_error.pklz")
            )
            exc.add_note(error_msg)
            if raise_errors or not job.result():
                raise exc
            else:
                logger.error("\nTask execution failed\n%s", error_msg)
        finally:
            self.run_start_time = None
        PersistentCache().clean_up()
        result = job.result()
        if result is None:
            if job.lockfile.exists():
                raise RuntimeError(
                    f"Job {job} has a lockfile, but no result was found. "
                    "This may be due to another submission that is currently running, or the hard "
                    "interrupt (e.g. a debugging abortion) interrupting a previous run. "
                    f"In the case of an interrupted run, please remove {str(job.lockfile)!r} "
                    "and resubmit."
                )
            raise RuntimeError(f"Job {job} has no result in {str(job.cache_dir)!r}")
        return result

    def submit(self, job: "Job[TaskType]", rerun: bool = False) -> None:
        """Submit a job to the worker.

        Parameters
        ----------
        job : :obj:`~pydra.engine.job.Job`
            The job to submit
        rerun : bool, optional
            Whether to force the re-computation of the job results even if existing
            results are found, by default False
        """
        if self.worker.is_async:  # Only workflow tasks can be async
            self.loop.run_until_complete(self.worker.submit(job, rerun=rerun))
        else:
            self.worker.run(job, rerun=rerun)

    def __getstate__(self):
        state = self.__dict__.copy()
        # Remove the unpicklable entries or those that should not be pickled
        # When unpickled (in another process) the submitter can't be called
        state["loop"] = None
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # Restore the loop and worker
        self.loop = get_open_loop()
        self.worker.loop = self.loop

    def expand_workflow(self, workflow_task: "Job[workflow.Task]", rerun: bool) -> None:
        """Expands and executes a workflow job synchronously. Typically only used during
        debugging and testing, as the asynchronous version is more efficient.

        Parameters
        ----------
        job : :obj:`~pydra.engine.job.Job[workflow.Task]`
            Workflow Job object

        """
        # Construct the workflow
        wf = workflow_task.task.construct()
        # Generate the execution graph
        exec_graph = wf.execution_graph(submitter=self)
        workflow_task.return_values = {"workflow": wf, "exec_graph": exec_graph}
        tasks = self.get_runnable_tasks(exec_graph)
        while tasks or any(not n.done for n in exec_graph.nodes):
            for job in tasks:
                self.worker.run(job, rerun=rerun and self.propagate_rerun)
            tasks = self.get_runnable_tasks(exec_graph)

    async def expand_workflow_async(
        self, workflow_task: "Job[workflow.Task]", rerun: bool
    ) -> None:
        """
        Expand and execute a workflow job asynchronously.

        Parameters
        ----------
        job : :obj:`~pydra.engine.job.Job[workflow.Task]`
            Workflow Job object
        """
        wf = workflow_task.task.construct()
        # Generate the execution graph
        exec_graph = wf.execution_graph(submitter=self)
        workflow_task.return_values = {"workflow": wf, "exec_graph": exec_graph}
        # keep track of pending futures
        task_futures = set()
        futured: dict[str, Job[TaskType]] = {}
        tasks = self.get_runnable_tasks(exec_graph)
        errors = []
        try:
            while tasks or task_futures or any(not n.done for n in exec_graph.nodes):
                if not tasks and not task_futures:
                    # it's possible that task_futures is empty, but not able to get any
                    # tasks from graph_copy (using get_runnable_tasks)
                    # this might be related to some delays saving the files
                    # so try to get_runnable_tasks for another minute
                    ii = 0
                    while not tasks and any(not n.done for n in exec_graph.nodes):
                        tasks = self.get_runnable_tasks(exec_graph)
                        ii += 1
                        # don't block the event loop!
                        await asyncio.sleep(1)
                        if ii > 10:
                            not_done = "\n".join(
                                (
                                    f"{n.name}: started={bool(n.started)}, "
                                    f"blocked={list(n.blocked)}, queued={list(n.queued)}"
                                )
                                for n in exec_graph.nodes
                                if not n.done
                            )
                            msg = (
                                "Something has gone wrong when retrieving the predecessor "
                                f"results. Not able to get any more tasks but he following "
                                f"nodes of the {wf.name!r} workflow are not done:"
                                f"\n{not_done}\n\n"
                            )
                            not_done = [n for n in exec_graph.nodes if not n.done]
                            msg += "\n" + ", ".join(
                                f"{t.name}: {t.done}"
                                for t in not_done[0].queued.values()
                            )
                            # Get blocked tasks and the predecessors they are blocked on
                            outstanding: dict[Job[TaskType], list[Job[TaskType]]] = {
                                t: [
                                    p
                                    for p in exec_graph.predecessors[t.name]
                                    if not p.done
                                ]
                                for t in exec_graph.sorted_nodes
                            }

                            hashes_have_changed = False
                            for job, blocked_on in outstanding.items():
                                if not blocked_on:
                                    continue
                                msg += f"- '{job.name}' node blocked due to\n"
                                for pred in blocked_on:
                                    if (
                                        pred.checksum
                                        != wf.inputs._graph_checksums[pred.name]
                                    ):
                                        msg += (
                                            f"    - hash changes in '{pred.name}' node "
                                            f"inputs. Current values and hashes: "
                                            f"{pred.inputs}, {pred.inputs._hashes}\n"
                                        )
                                        hashes_have_changed = True
                                    elif pred not in outstanding:
                                        msg += (
                                            f"    - undiagnosed issues in '{pred.name}' "
                                            "node, potentially related to file-system "
                                            "access issues "
                                        )
                                msg += "\n"
                            if hashes_have_changed:
                                msg += (
                                    "Set loglevel to 'debug' in order to track hash "
                                    "changes throughout the execution of the workflow."
                                    "\n\n These issues may have been caused by "
                                    "`bytes_repr()` methods that don't return stable "
                                    "hash values for specific object types across "
                                    "multiple processes (see bytes_repr() "
                                    '"singledispatch "function in pydra/utils/hash.py).'
                                    "You may need to write specific `bytes_repr()` "
                                    "implementations (see `pydra.utils.hash.register_serializer`) "
                                    "or `__bytes_repr__()` dunder methods to handle one "
                                    "or more types in your interface inputs."
                                )
                            raise RuntimeError(msg)
                for job in tasks:
                    if job.is_async:  # Only workflows at this stage
                        await self.worker.submit(
                            job, rerun=rerun and self.propagate_rerun
                        )
                    elif job.checksum not in futured:
                        asyncio_task = asyncio.Task(
                            self.worker.run(job, rerun=rerun and self.propagate_rerun),
                            name=job.checksum,
                        )
                        task_futures.add(asyncio_task)
                        futured[job.checksum] = job
                task_futures, completed = await self.fetch_finished(task_futures)
                for task_future in completed:
                    try:
                        task_future.result()
                    except Exception:
                        error_msg = format_exc()
                        if match := re.match(
                            r'.*"""(.*)""".*',
                            error_msg,
                            flags=re.DOTALL | re.MULTILINE,
                        ):
                            error_msg = match.group(1)
                        job = futured[task_future.get_name()]
                        task_name = job.name
                        if job.state_index is not None:
                            task_name += f"({job.state_index})"
                        errors.append(
                            f"Job {task_name!r}, {job.task!r}, errored:{error_msg}"
                        )
                tasks = self.get_runnable_tasks(exec_graph)
        finally:
            if errors:
                all_errors = "\n\n".join(errors)
                raise RuntimeError(
                    f"Workflow job {workflow_task} failed with errors"
                    f":\n\n{all_errors}\n\nSee output directory for details: {workflow_task.cache_dir}"
                )

    async def fetch_finished(
        self, futures
    ) -> tuple[set[asyncio.Task], set[asyncio.Task]]:
        """
        Awaits asyncio's :class:`asyncio.Task` until one is finished.

        Parameters
        ----------
        futures : set of asyncio awaitables
            Job execution coroutines or asyncio :class:`asyncio.Task`

        Returns
        -------
        pending : set
            Pending asyncio :class:`asyncio.Task`.
        done: set
            Completed asyncio :class:`asyncio.Task`

        """
        done = set()
        try:
            done, pending = await asyncio.wait(
                futures, return_when=asyncio.FIRST_COMPLETED
            )
        except ValueError:
            # nothing pending!
            pending = set()
        logger.debug(f"Tasks finished: {len(done)}")
        return pending, done

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        """
        Close submitter.

        Do not close previously queued loop.

        """
        self.worker.close()
        if self._own_loop:
            self.loop.close()

    def _check_locks(self, tasks: list[Job]) -> None:
        """Check for stale lock files and remove them."""
        if self.clean_stale_locks:
            for job in tasks:
                start_time = job.run_start_time
                if start_time and start_time < self.run_start_time:
                    job.lockfile.unlink()

    def get_runnable_tasks(self, graph: DiGraph) -> list["Job[TaskType]"]:
        """Parse a graph and return all runnable tasks.

        Parameters
        ----------
        graph : :obj:`~pydra.engine.graph.DiGraph`
            Graph object

        Returns
        -------
        tasks : list of :obj:`~pydra.engine.job.Job`
            List of runnable tasks
        following_err : dict[NodeToExecute, list[str]]
            Dictionary of tasks that are blocked by errored tasks
        """
        tasks = []
        not_started = set()
        node: NodeExecution
        for node in graph.sorted_nodes:
            if node.done:
                continue
            # since the list is sorted (breadth-first) we can stop
            # when we find a job that depends on any job that is already in tasks
            preds = set(graph.predecessors[node.name])
            if preds.intersection(not_started):
                break
            # Record if the node has not been started
            if not node.started:
                not_started.add(node)
            tasks.extend(node.get_runnable_tasks(graph))
        self._check_locks(tasks)
        if len(tasks) > self.max_concurrent:
            logger.info(
                "Reducing number of tasks to run concurrently from %d to %d",
                len(tasks),
                self.max_concurrent,
            )
            tasks = tasks[: self.max_concurrent]
        return tasks

    @property
    def cache_root(self):
        """Get the location of the cache directory."""
        return self._cache_root

    @cache_root.setter
    def cache_root(self, location):
        if location is not None:
            self._cache_root = Path(location).resolve()
            self._cache_root.mkdir(parents=False, exist_ok=True)
        else:
            self._cache_root = mkdtemp()
            self._cache_root = Path(self._cache_root).resolve()


class NodeExecution(ty.Generic[TaskType]):
    """A wrapper around a workflow node containing the execution state of the tasks that
    are generated from it"""

    name: str
    node: "Node"
    submitter: Submitter

    # List of tasks that were completed successfully
    successful: dict[int, list["Job[TaskType]"]]
    # List of tasks that failed
    errored: dict[int, "Job[TaskType]"]
    # List of tasks that couldn't be run due to upstream errors
    unrunnable: dict[int, list["Job[TaskType]"]]
    # List of tasks that are queued
    queued: dict[int, "Job[TaskType]"]
    # List of tasks that are queued
    running: dict[int, tuple["Job[TaskType]", datetime]]
    # List of tasks that are blocked on other tasks to complete before they can be run
    blocked: dict[int, "Job[TaskType]"] | None

    _tasks: dict[int | None, "Job[TaskType]"] | None

    workflow: "Workflow"

    graph: DiGraph["NodeExecution"] | None

    def __init__(
        self,
        node: "Node",
        submitter: Submitter,
        workflow: "Workflow",
    ):
        self.name = node.name
        self.node = node
        self.submitter = submitter
        # Initialize the state dictionaries
        self._tasks = None
        self.blocked = None
        self.successful = {}
        self.errored = {}
        self.queued = {}
        self.running = {}  # Not used in logic, but may be useful for progress tracking
        self.unrunnable = defaultdict(list)
        self.workflow = workflow
        self.graph = None

    @property
    def state(self):
        return self.node.state

    def __repr__(self):
        return (
            f"NodeExecution(name={self.name!r}, blocked={list(self.blocked)}, "
            f"queued={list(self.queued)}, running={list(self.running)}, "
            f"successful={list(self.successful)}, errored={list(self.errored)}, "
            f"unrunnable={list(self.unrunnable)})"
        )

    @property
    def inputs(self) -> "Node.Inputs":
        return self.node.inputs

    @property
    def _task(self) -> "Node":
        return self.node._task

    @property
    def tasks(self) -> ty.Generator["Job[TaskType]", None, None]:
        if self._tasks is None:
            raise RuntimeError("Tasks have not been generated")
        return self._tasks.values()

    def start(self) -> None:
        """Prepare the execution node so that it can be processed"""
        self._tasks = {}
        if self.state:
            values = {}
            for name, value in self.node.state_values.items():
                if name in self.node.state.current_splitter_rpn:
                    if name in ("*", "."):
                        continue
                    if isinstance(value, LazyField):
                        values[name] = value._get_value(
                            workflow=self.workflow, graph=self.graph
                        )
                    else:
                        values[name] = value
            self.state.prepare_states(values)
            self.state.prepare_inputs()
            # Generate the tasks
            for index, split_defn in enumerate(self._split_task()):
                self._tasks[index] = Job(
                    task=split_defn,
                    submitter=self.submitter,
                    environment=self.node._environment,
                    name=self.node.name,
                    hooks=self.node._hooks,
                    state_index=index,
                )
        else:
            self._tasks[None] = Job(
                task=self._resolve_lazy_inputs(task=self.node._task),
                submitter=self.submitter,
                environment=self.node._environment,
                hooks=self.node._hooks,
                name=self.node.name,
            )
        self.blocked = copy(self._tasks)

    @property
    def started(self) -> bool:
        return (
            self.successful
            or self.errored
            or self.unrunnable
            or self.queued
            or self.blocked is not None
        )

    @property
    def done(self) -> bool:
        self.update_status()
        if not self.started:
            return False
        # Check to see if any previously queued tasks have completed
        return not (self.queued or self.blocked or self.running)

    @property
    def has_errored(self) -> bool:
        self.update_status()
        return bool(self.errored)

    def update_status(self) -> None:
        """Updates the status of the tasks in the node."""
        if not self.started:
            return
        # Check to see if any previously queued tasks have completed
        for index, job in list(self.queued.items()):
            try:
                is_done = job.done
            except ValueError:
                errored = True
                is_done = False
            else:
                errored = False
            if is_done:
                self.successful[job.state_index] = self.queued.pop(index)
            elif job.errored or errored:
                self.errored[job.state_index] = self.queued.pop(index)
            elif job.run_start_time:
                self.running[job.state_index] = (
                    self.queued.pop(index),
                    job.run_start_time,
                )
        # Check to see if any previously running tasks have completed
        for index, (job, _) in list(self.running.items()):
            if job.done:
                self.successful[job.state_index] = self.running.pop(index)[0]
            elif job.errored:
                self.errored[job.state_index] = self.running.pop(index)[0]

    @property
    def all_failed(self) -> bool:
        return (self.unrunnable or self.errored) and not (
            self.successful or self.blocked or self.queued
        )

    def _resolve_lazy_inputs(
        self,
        task: "Task",
        state_index: int | None = None,
    ) -> "Task":
        """Resolves lazy fields in the task by replacing them with their
        actual values calculated by upstream jobs.

        Parameters
        ----------
        task : Task
            The task to resolve the lazy fields of
        state_index : int, optional
            The state index for the workflow, by default None

        Returns
        -------
        Task
            The task with all lazy fields resolved
        """
        resolved = {}
        for name, value in attrs_values(task).items():
            if isinstance(value, LazyField):
                resolved[name] = value._get_value(
                    workflow=self.workflow, graph=self.graph, state_index=state_index
                )
        return attrs.evolve(task, **resolved)

    def _split_task(self) -> dict[int, "Task[OutputType]"]:
        """Split the task into the different states it will be run over

        Parameters
        ----------
        values : dict[str, Any]
            The values to use for the split
        """
        # TODO: doesn't work properly for more cmplicated wf (check if still an issue)
        if not self.node.state:
            return {None: self.node._task}
        split_defs = []
        for index, vals in zip(self.node.state.inputs_ind, self.node.state.states_val):
            resolved = {}
            for inpt_name in set(self.node.input_names):
                value = getattr(self._task, inpt_name)
                state_key = f"{self.node.name}.{inpt_name}"
                try:
                    resolved[inpt_name] = vals[state_key]
                except KeyError:
                    if isinstance(value, LazyField):
                        resolved[inpt_name] = value._get_value(
                            workflow=self.workflow,
                            graph=self.graph,
                            state_index=index.get(state_key),
                        )
            split_defs.append(attrs.evolve(self.node._task, **resolved))
        return split_defs

    def get_runnable_tasks(self, graph: DiGraph) -> list["Job[TaskType]"]:
        """For a given node, check to see which tasks have been successfully run, are ready
        to run, can't be run due to upstream errors, or are blocked on other tasks to complete.

        Parameters
        ----------
        node : :obj:`~pydra.engine.node.Node`
            The node object to get the tasks for
        graph : :obj:`~pydra.engine.graph.DiGraph`
            Graph object


        Returns
        -------
        runnable : list[NodeExecution]
            List of tasks that are ready to run
        """
        runnable: list["Job[TaskType]"] = []
        predecessors: list["Job[TaskType]"] = graph.predecessors[self.node.name]

        # If there is a split, we need to wait for all predecessor nodes to finish
        # In theory, if the current splitter splits an already split state we should
        # only need to wait for the direct predecessor jobs to finish, however, this
        # would require a deep refactor of the State class as we need the whole state
        # in order to assign consistent state indices across the new split

        # FIXME: The branch for handling partially completed/errored/unrunnable
        # predecessor nodes can't be used until the State class can be partially
        # initialised with lazy-fields.
        if True:  # self.node.splitter:
            if unrunnable := [p for p in predecessors if p.errored or p.unrunnable]:
                self.unrunnable = {None: unrunnable}
                self.blocked = {}
                assert self.done
            else:
                if all(p.done for p in predecessors):
                    if not self.started:
                        self.start()
                    if self.node.state is None:
                        inds = [None]
                    else:
                        inds = list(range(len(self.node.state.states_ind)))
                    if self.blocked:
                        for i in inds:
                            runnable.append(self.blocked.pop(i))
        else:
            if not self.started:
                self.start()

            # Check to see if any blocked tasks are now runnable/unrunnable
            for index, job in list(self.blocked.items()):
                pred: NodeExecution
                is_runnable = True
                states_ind = (
                    list(self.node.state.states_ind[index].items())
                    if self.node.state
                    else []
                )
                for pred in predecessors:
                    if pred.node.state:
                        pred_states_ind = {
                            (k, i)
                            for k, i in states_ind
                            if k.startswith(pred.name + ".")
                        }
                        pred_inds = [
                            i
                            for i, ind in enumerate(pred.node.state.states_ind)
                            if set(ind.items()).issuperset(pred_states_ind)
                        ]
                    else:
                        pred_inds = [None]
                    if not all(i in pred.successful for i in pred_inds):
                        is_runnable = False
                        blocked = True
                        if pred_errored := [
                            pred.errored[i] for i in pred_inds if i in pred.errored
                        ]:
                            self.unrunnable[index].extend(pred_errored)
                            blocked = False
                        if pred_unrunnable := [
                            pred.unrunnable[i]
                            for i in pred_inds
                            if i in pred.unrunnable
                        ]:
                            self.unrunnable[index].extend(pred_unrunnable)
                            blocked = False
                        if not blocked:
                            del self.blocked[index]
                        break
                if is_runnable:
                    runnable.append(self.blocked.pop(index))
        self.queued.update({t.state_index: t for t in runnable})
        return list(self.queued.values())


async def prepare_runnable(runnable):
    return runnable.pickle_task()


# def _list_blocked_tasks(graph):
#     """Generates a list of tasks that can't be run and predecessors that are blocking
#     them to help debugging of broken workflows"""
#     blocked = []
#     for tsk in graph.sorted_nodes:
#         blocking = []
#         for pred in graph.predecessors[tsk.name]:
#             if not pred.done:
#                 matching_name = []
#                 for cache_loc in tsk.readonly_caches:
#                     for tsk_work_dir in cache_loc.iterdir():
#                         if (tsk_work_dir / "_job.pklz").exists():
#                             with open(tsk_work_dir / "_job.pklz", "rb") as f:
#                                 saved_tsk = pickle.load(f)
#                             if saved_tsk.name == pred.name:
#                                 matching_name.append(
#                                     f"{saved_tsk.name} ({tsk_work_dir.name})"
#                                 )
#                 blocking.append((pred, ", ".join(matching_name)))
#         if blocking:
#             blocked.append(
#                 f"\n{tsk.name} ({tsk.checksum}) is blocked by "
#                 + "; ".join(
#                     f"{pred.name} ({pred.checksum}), which matches names of [{matching}]"
#                     for pred, matching in blocking
#                 )
#             )
#     return blocked


def get_open_loop():
    """
    Get current event loop.

    If the loop is closed, a new
    loop is created and set as the current event loop.

    Returns
    -------
    loop : :obj:`asyncio.EventLoop`
        The current event loop

    """
    if os.name == "nt":
        loop = asyncio.ProactorEventLoop()  # for subprocess' pipes on Windows
    else:
        try:
            loop = asyncio.get_event_loop()
        # in case RuntimeError: There is no current event loop in thread 'MainThread'
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        else:
            if loop.is_closed():
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
    return loop



"""Decorators and helper functions to create ShellTasks used in Pydra workflows"""

from __future__ import annotations
import typing as ty
import re
import glob
from collections import defaultdict
import inspect
from copy import copy
import attrs
import builtins
from typing import dataclass_transform
from fileformats.core import from_mime
from fileformats import generic
from fileformats.core.exceptions import FormatRecognitionError
from pydra.utils.general import attrs_values
from pydra.compose.base import (
    Arg,
    Out,
    check_explicit_fields_are_none,
    extract_fields_from_class,
    ensure_field_objects,
    build_task_class,
    sanitize_xor,
    NO_DEFAULT,
)
from pydra.utils.typing import (
    is_fileset_or_union,
    MultiInputObj,
    TypeParser,
    is_optional,
)
from . import field
from .task import Task, Outputs


@dataclass_transform(
    kw_only_default=True,
    field_specifiers=(field.out, field.outarg),
)
def outputs(wrapped):
    """Decorator to specify the output fields of a shell command is a dataclass-style type"""
    return wrapped


@dataclass_transform(
    kw_only_default=True,
    field_specifiers=(field.arg,),
)
def define(
    wrapped: type | str | None = None,
    /,
    inputs: list[str | Arg] | dict[str, Arg | type] | None = None,
    outputs: list[str | Out] | dict[str, Out | type] | type | None = None,
    bases: ty.Sequence[type] = (),
    outputs_bases: ty.Sequence[type] = (),
    auto_attribs: bool = True,
    name: str | None = None,
    xor: ty.Sequence[str | None] | ty.Sequence[ty.Sequence[str | None]] = (),
) -> "Task":
    """Create a task for a shell command. Can be used either as a decorator on
    the "canonical" dataclass-form of a task or as a function that takes a
    "shell-command template string" of the form

    ```
    shell.define("command <input1> <input2> --output <out|output1>")
    ```

    Fields are inferred from the template if not provided. In the template, inputs are
    specified with `<fieldname>` and outputs with `<out|fieldname>`.

    ```
    my_command <myinput> <out|myoutput2>
    ```

    The types of the fields can be specified using their MIME like (see fileformats.core.from_mime), e.g.

    ```
    my_command <myinput:text/csv> <out|myoutput2:image/png>
    ```

    The template can also specify options with `-` or `--` followed by the option name
    and arguments with `<argname:type>`. The type is optional and will default to
    `generic/fs-object` if not provided for arguments and `field/text` for
    options. The file-formats namespace can be dropped for generic and field formats, e.g.

    ```
    another-command <input1:directory> <input2:int> --output <out|output1:text/csv>
    ```

    Parameters
    ----------
    wrapped : type | str | None
        The class or command line template to create an interface for
    inputs : list[str | Arg] | dict[str, Arg | type] | None
        The input fields of the shell command
    outputs : list[str | Out] | dict[str, Out | type] | type | None
        The output fields of the shell command
    auto_attribs : bool
        Whether to use auto_attribs mode when creating the class
    args_last : bool
        Whether to put the executable argument last in the command line instead of first
        as they appear in the template
    name: str | None
        The name of the returned class
    xor: Sequence[str | None] | Sequence[Sequence[str | None]], optional
        Names of args that are exclusive mutually exclusive, which must include
        the name of the current field. If this list includes None, then none of the
        fields need to be set.

    Returns
    -------
    Task
        The interface for the shell command
    """

    def make(
        wrapped: ty.Callable | type | None = None,
    ) -> Task:

        if inspect.isclass(wrapped):
            klass = wrapped
            executable: str
            try:
                executable = attrs.fields(klass).executable.default
            except (AttributeError, attrs.exceptions.NotAnAttrsClassError):
                try:
                    executable = klass.executable
                except AttributeError:
                    raise AttributeError(
                        f"Shell task class {wrapped} must have an `executable` "
                        "attribute that specifies the command to run"
                    ) from None
            if not isinstance(executable, str) and not (
                isinstance(executable, ty.Sequence)
                and all(isinstance(e, str) for e in executable)
            ):
                raise ValueError(
                    "executable must be a string or a sequence of strings"
                    f", not {executable!r}"
                )
            class_name = klass.__name__
            check_explicit_fields_are_none(klass, inputs, outputs)
            parsed_inputs, parsed_outputs = extract_fields_from_class(
                Task,
                Outputs,
                klass,
                field.arg,
                field.out,
                auto_attribs,
                skip_fields=["executable"],
            )
        else:
            if not isinstance(wrapped, (str, list)):
                raise ValueError(
                    f"wrapped must be a class or a string, not {wrapped!r}"
                )
            klass = None
            input_helps, output_helps = {}, {}

            executable, inferred_inputs, inferred_outputs = parse_command_line_template(
                wrapped,
                inputs=inputs,
                outputs=outputs,
            )

            parsed_inputs, parsed_outputs = ensure_field_objects(
                arg_type=field.arg,
                out_type=field.out,
                inputs=inferred_inputs,
                outputs=inferred_outputs,
                input_helps=input_helps,
                output_helps=output_helps,
            )

            if name:
                class_name = name
            else:
                class_name = (
                    "_".join(executable) if isinstance(executable, list) else executable
                )
                class_name = re.sub(r"[^\w]", "_", class_name)
            if class_name[0].isdigit():
                class_name = f"_{class_name}"

            # Add in fields from base classes
            parsed_inputs.update({n: getattr(Task, n) for n in Task.BASE_ATTRS})
            parsed_outputs.update({n: getattr(Outputs, n) for n in Outputs.BASE_ATTRS})

        if "executable" in parsed_inputs:
            raise ValueError(
                "The argument 'executable' is reserved for a field to hold the command "
                "to be run"
            )

        # Update the inputs (overriding inputs from base classes) with the executable
        # and the output argument fields
        parsed_inputs.update(
            {o.name: o for o in parsed_outputs.values() if isinstance(o, field.arg)}
        )
        parsed_inputs["executable"] = field.arg(
            name="executable",
            type=str | ty.Sequence[str],
            argstr="",
            position=0,
            default=executable,
            validator=attrs.validators.min_len(1),
            help=Task.EXECUTABLE_HELP,
        )

        # Set positions for the remaining inputs that don't have an explicit position
        position_stack = remaining_positions(list(parsed_inputs.values()), xor=xor)
        for inpt in parsed_inputs.values():
            if inpt.name == "append_args":
                continue
            if inpt.position is None:
                inpt.position = position_stack.pop(0)

        # Convert string default values to callables that glob the files in the cwd
        for outpt in parsed_outputs.values():
            if (
                isinstance(outpt, field.out)
                and isinstance(outpt.default, str)
                and TypeParser.contains_type(generic.FileSet, outpt.type)
            ):
                outpt.callable = GlobCallable(outpt.default)
                outpt.default = NO_DEFAULT

        defn = build_task_class(
            Task,
            Outputs,
            parsed_inputs,
            parsed_outputs,
            name=class_name,
            klass=klass,
            bases=bases,
            outputs_bases=outputs_bases,
            xor=xor,
        )
        return defn

    # If a name is provided (and hence not being used as a decorator), check to see if
    # we are extending from a class that already defines an executable
    if wrapped is None and name is not None:
        for base in bases:
            try:
                wrapped = attrs.fields(base).executable.default
            except (AttributeError, attrs.exceptions.NotAnAttrsClassError):
                try:
                    wrapped = base.executable
                except AttributeError:
                    pass
            if wrapped:
                break
        if wrapped is None:
            raise ValueError(
                f"name ({name!r}) can only be provided when creating a class "
                "dynamically, i.e. not using it as a decorator. Check to see "
                "whether you have forgotten to provide the command line template"
            )
    # If wrapped is provided (i.e. this is not being used as a decorator), return the
    # interface class
    if wrapped is not None:
        if not isinstance(wrapped, (type, str, list)):
            raise ValueError(
                f"wrapped must be a class, a string or a list, not {wrapped!r}"
            )
        return make(wrapped)
    return make


def parse_command_line_template(
    template: str,
    inputs: list[str | Arg] | dict[str, Arg | type] | None = None,
    outputs: list[str | Out] | dict[str, Out | type] | None = None,
) -> ty.Tuple[str, dict[str, Arg | type], dict[str, Out | type]]:
    """Parses a command line template into a name and input and output fields. Fields
    are inferred from the template if not explicitly provided.

    In the template, inputs are specified with `<fieldname>` and outputs with `<out|fieldname>`.
    The types of the fields can be specified using their MIME like (see fileformats.core.from_mime), e.g.

    ```
    my_command <myinput> <out|myoutput2>
    ```

    The template can also specify options with `-` or `--`
    followed by the option name and arguments with `<argname:type>`. The type is optional and
    will default to `generic/fs-object` if not provided for arguments and `field/text` for
    options. The file-formats namespace can be dropped for generic and field formats, e.g.

    ```
    another-command <input1:directory> <input2:int> --output <out|output1:text/csv>
    ```

    Parameters
    ----------
    template : str
        The command line template
    inputs : list[str | Arg] | dict[str, Arg | type] | None
        The input fields of the shell command
    outputs : list[str | Out] | dict[str, Out | type] | type | None
        The output fields of the shell command

    Returns
    -------
    executable : str
        The name of the command line template
    inputs : dict[str, Arg | type]
        The input fields of the command line template
    outputs : dict[str, Out | type]
        The output fields of the command line template

    Raises
    ------
    ValueError
        If an unknown token is found in the command line template
    TypeError
        If an unknown type is found in the command line template
    """
    if isinstance(inputs, list):
        inputs = {arg.name: arg for arg in inputs}
    elif isinstance(inputs, dict):
        inputs = copy(inputs)  # We don't want to modify the original
    else:
        assert inputs is None
        inputs = {}
    if isinstance(outputs, list):
        outputs = {o.name: o for o in outputs}
    elif isinstance(outputs, dict):
        outputs = copy(outputs)  # We don't want to modify the original
    else:
        assert outputs is None
        outputs = {}
    if isinstance(template, list):
        tokens = template
    else:
        tokens = template.split()
    executable = []
    start_args_index = 0
    for part in tokens:
        if part.startswith("<") or part.startswith("-"):
            break
        executable.append(part)
        start_args_index += 1
    if not executable:
        raise ValueError(f"Found no executable in command line template: {template}")
    if len(executable) == 1:
        executable = executable[0]
    tokens = tokens[start_args_index:]
    if not tokens:
        return executable, inputs, outputs
    arg_pattern = r"<([:a-zA-Z0-9_,\|\-\.\/\+\*]+(?:\?|(?:=|\$)[^>]+)?)>"
    opt_pattern = r"--?[a-zA-Z0-9_\-]+"
    arg_re = re.compile(arg_pattern)
    opt_re = re.compile(opt_pattern)
    bool_arg_re = re.compile(f"({opt_pattern}){arg_pattern}")

    arguments = []
    option = None

    def add_arg(name, field_type, kwds):
        """Merge the typing information with an existing field if it exists"""
        if issubclass(field_type, Out):
            dct = outputs
        else:
            dct = inputs
        try:
            fld = dct.pop(name)
        except KeyError:
            fld = field_type(name=name, **kwds)
        else:
            if isinstance(fld, dict):
                fld = field_type(**fld)
            elif isinstance(fld, type) or ty.get_origin(fld):
                kwds["type"] = fld
                fld = field_type(name=name, **kwds)
            elif not isinstance(fld, field_type):  # If fld type is outarg not out
                fld = field_type(**attrs_values(fld))
            fld.name = name
            type_ = kwds.pop("type", fld.type)
            if fld.type is ty.Any:
                fld.type = type_
            for k, v in kwds.items():
                setattr(fld, k, v)
        dct[name] = fld
        if issubclass(field_type, Arg):
            arguments.append(fld)

    def from_type_str(type_str) -> type:
        types = []
        for tp in type_str.split(","):
            if "/" in tp:
                type_ = from_mime(tp)
            elif tp == "...":
                type_ = "..."
            else:
                if tp in ("int", "float", "str", "bool"):
                    type_ = getattr(builtins, tp)
                else:
                    try:
                        type_ = from_mime(f"generic/{tp}")
                    except FormatRecognitionError:
                        raise TypeError(
                            f"Found unknown type, {tp!r}, in command template: {template!r}"
                        ) from None
            types.append(type_)
        if len(types) == 2 and types[1] == "...":
            type_ = tuple[types[0], ...]
        elif len(types) > 1:
            type_ = tuple[*types]
        else:
            type_ = types[0]
        return type_

    for token in tokens:
        if match := arg_re.match(token):
            name = match.group(1)
            modify = False
            if name.startswith("out|"):
                name = name[4:]
                field_type = field.outarg
            elif name.startswith("modify|"):
                name = name[7:]
                field_type = field.arg
                modify = True
            else:
                field_type = field.arg
            # Identify type after ':' symbols
            kwds = {}
            is_multi = False
            optional = False
            if name.endswith("?"):
                assert "=" not in name
                name = name[:-1]
                optional = True
                kwds["default"] = None
            elif name.endswith("+"):
                is_multi = True
                name = name[:-1]
            elif name.endswith("*"):
                is_multi = True
                name = name[:-1]
                kwds["default"] = attrs.Factory(list)
            elif "=" in name:
                name, default = name.split("=")
                kwds["default"] = (
                    default[1:-1] if re.match(r"('|\").*\1", default) else eval(default)
                )
            elif "$" in name:
                name, path_template = name.split("$")
                kwds["path_template"] = path_template
                if field_type is not field.outarg:
                    raise ValueError(
                        f"Path templates can only be used with output fields, not {token}"
                    )
            if ":" in name:
                name, type_str = name.split(":")
                type_ = from_type_str(type_str)
                if ty.get_origin(type_) is tuple:
                    kwds["sep"] = " "
            else:
                type_ = generic.FsObject if option is None else str
            if is_multi:
                type_ = MultiInputObj[type_]
            if optional:
                type_ |= None  # Make the arguments optional
            kwds["type"] = type_
            if modify:
                kwds["copy_mode"] = generic.File.CopyMode.copy
                # Add field to outputs with the same name as the input
                add_arg(
                    name,
                    field.out,
                    {"type": type_, "callable": _InputPassThrough(name)},
                )
            # If name contains a '.', treat it as a file template and strip it from the name
            if field_type is field.outarg and "path_template" not in kwds:
                path_template = name
                if is_fileset_or_union(type_):
                    if ty.get_origin(type_):
                        ext_type = next(a for a in ty.get_args(type_) if a is not None)
                    else:
                        ext_type = type_
                    if ext_type.ext is not None:
                        path_template = name + ext_type.ext
                kwds["path_template"] = path_template
            # Set the default value to None if the field is optional and no default is
            # provided
            if is_optional(type_) and "default" not in kwds:
                kwds["default"] = None
            if option is None:
                add_arg(name, field_type, kwds)
            else:
                kwds["argstr"] = option
                add_arg(name, field_type, kwds)
                option = None

        elif match := bool_arg_re.match(token):
            argstr, var = match.groups()
            if "=" in var:
                var, default = var.split("=")
                default = eval(default)
            else:
                default = False
            add_arg(
                var, field.arg, {"type": bool, "argstr": argstr, "default": default}
            )
            option = None
        elif match := opt_re.match(token):
            option = token
        else:
            raise ValueError(
                f"Found unknown token {token!r} in command line template: {template}"
            )

    if option:
        raise ValueError(f"Found an option without a field: {option!r}")

    remaining_pos = remaining_positions(arguments, len(arguments) + 1, 1)

    for argument in arguments:
        if argument.position is None:
            argument.position = remaining_pos.pop(0)

    return executable, inputs, outputs


def remaining_positions(
    args: list[Arg],
    num_args: int | None = None,
    start: int = 0,
    xor: set[frozenset[str]] | None = None,
) -> ty.List[int]:
    """Get the remaining positions for input fields

    Parameters
    ----------
    args : list[Arg]
        The list of input fields
    num_args : int, optional
        The number of arguments, by default it is the length of the args
    start : int, optional
        The starting position, by default 0
    xor : set[frozenset[str]], optional
        A set of mutually exclusive fields, by default None

    Returns
    -------
    list[int]
        The list of remaining positions

    Raises
    ------
    ValueError
        If multiple fields have the same position
    """
    xor = sanitize_xor(xor)
    if num_args is None:
        num_args = len(args) - 1  # Subtract 1 for the 'append_args' field
    # Check for multiple positions
    positions = defaultdict(list)
    for arg in args:
        if arg.name == "append_args":
            continue
        if arg.position is not None:
            if arg.position >= 0:
                positions[arg.position].append(arg)
            else:
                positions[num_args + arg.position].append(arg)
    if multiple_positions := {
        k: [f"{a.name}({a.position})" for a in v]
        for k, v in positions.items()
        if len(v) > 1 and not any(x.issuperset(a.name for a in v) for x in xor)
    }:
        raise ValueError(
            f"Multiple fields have the overlapping positions: {multiple_positions}"
        )
    return [i for i in range(start, num_args) if i not in positions]


@attrs.define
class _InputPassThrough:
    """A class that can be used to pass through an input to the output"""

    name: str

    def __call__(self, inputs: Task) -> ty.Any:
        return getattr(inputs, self.name)


class GlobCallable:
    """Callable that can be used to glob files"""

    def __init__(self, pattern: str):
        self.pattern = pattern

    def __call__(self) -> generic.FileSet:
        matches = glob.glob(self.pattern)
        if not matches:
            raise FileNotFoundError(f"No files found matching pattern: {self.pattern}")
        return matches



import typing as ty
import abc
import attrs
from typing import Self
from pydra.utils.typing import StateArray
from pydra.utils.hash import hash_single
from pydra.engine import node

if ty.TYPE_CHECKING:
    from pydra.engine.submitter import DiGraph, NodeExecution
    from pydra.engine.job import Job
    from pydra.engine.workflow import Workflow
    from pydra.compose.base import Task


T = ty.TypeVar("T")
TaskType = ty.TypeVar("TaskType", bound="Task")

TypeOrAny = ty.Union[type, ty.Any]


@attrs.define(kw_only=True)
class LazyField(ty.Generic[T], metaclass=abc.ABCMeta):
    """Lazy fields implement promises."""

    _field: str
    _type: TypeOrAny
    _cast_from: ty.Optional[ty.Type[ty.Any]] = None
    _type_checked: bool = False

    def __bytes_repr__(self, cache):
        yield type(self).__name__.encode() + b"("
        yield b"source=" + bytes(hash_single(self._source, cache))
        yield b"field=" + self._field.encode()
        yield b"type=" + bytes(hash_single(self._type, cache))
        yield b"cast_from=" + bytes(hash_single(self._cast_from, cache))
        yield b")"

    def _apply_cast(self, value):
        """\"Casts\" the value from the retrieved type if a cast has been applied to
        the lazy-field"""
        from pydra.utils.typing import TypeParser

        if self._cast_from:
            assert TypeParser.matches(value, self._cast_from)
            value = self._type(value)
        return value

    def _get_value(
        self,
        workflow: "Workflow",
        graph: "DiGraph[NodeExecution]",
        state_index: int | None = None,
    ) -> ty.Any:
        """Return the value of a lazy field.

        Parameters
        ----------
        workflow: Workflow
            the workflow object
        graph: DiGraph[NodeExecution]
            the graph representing the execution state of the workflow
        state_index : int, optional
            the state index of the field to access

        Returns
        -------
        value : Any
            the resolved value of the lazy-field
        """
        raise NotImplementedError("LazyField is an abstract class")

    def split(self) -> Self:
        """ "Splits" the lazy field over an array of nodes by replacing the sequence type
        of the lazy field with StateArray to signify that it will be "split" across
        """
        from pydra.utils.typing import (
            TypeParser,
        )  # pylint: disable=import-outside-toplevel

        # Modify the type of the lazy field to include the split across a state-array
        inner_type, prev_split_depth = TypeParser.strip_splits(self._type)
        assert prev_split_depth <= 1
        if inner_type is ty.Any:
            type_ = StateArray[ty.Any]
        elif TypeParser.matches_type(inner_type, list):
            item_type = TypeParser.get_item_type(inner_type)
            type_ = StateArray[item_type]
        else:
            raise TypeError(
                f"Cannot split non-sequence field {self}  of type {inner_type}"
            )
        if prev_split_depth:
            type_ = StateArray[
                type_
            ]  # FIXME: This nesting of StateArray is probably unnecessary
        return attrs.evolve(self, type=type_)


@attrs.define(kw_only=True)
class LazyInField(LazyField[T]):

    _workflow: "Workflow" = attrs.field()

    _attr_type = "input"

    def __eq__(self, other):
        return (
            isinstance(other, LazyInField)
            and self._field == other._field
            and self._type == other._type
        )

    def __repr__(self):
        return f"{type(self).__name__}(field={self._field!r}, type={self._type})"

    @property
    def _source(self):
        return self._workflow

    def _get_value(
        self,
        workflow: "Workflow",
        graph: "DiGraph[NodeExecution]",
        state_index: int | None = None,
    ) -> ty.Any:
        """Return the value of a lazy field.

        Parameters
        ----------
        workflow: Workflow
            the workflow object
        graph: DiGraph[NodeExecution]
            the graph representing the execution state of the workflow
        state_index : int, optional
            the state index of the field to access

        Returns
        -------
        value : Any
            the resolved value of the lazy-field
        """
        value = workflow.inputs[self._field]
        value = self._apply_cast(value)
        return value


@attrs.define(kw_only=True)
class LazyOutField(LazyField[T]):

    _node: node.Node
    _attr_type = "output"

    def __repr__(self):
        return (
            f"{type(self).__name__}(node={self._node.name!r}, "
            f"field={self._field!r}, type={self._type})"
        )

    def _get_value(
        self,
        workflow: "Workflow",
        graph: "DiGraph[NodeExecution]",
        state_index: int | None = None,
    ) -> ty.Any:
        """Return the value of a lazy field.

        Parameters
        ----------
        workflow: Workflow
            the workflow object
        graph: DiGraph[NodeExecution]
            the graph representing the execution state of the workflow
        state_index : int, optional
            the state index of the field to access

        Returns
        -------
        value : Any
            the resolved value of the lazy-field
        """

        def retrieve_from_job(job: "Job[TaskType]") -> ty.Any:
            if job.errored:
                raise ValueError(
                    f"Cannot retrieve value for {self._field!r} from {self._node.name} as "
                    "the node errored"
                )
            res = job.result()
            if res is None:
                raise RuntimeError(
                    f"Could not find results of '{job.name}' node in a sub-directory "
                    f"named '{{{job.checksum}}}' in any of the cache locations.\n"
                    + "\n".join(str(p) for p in set(job.readonly_caches))
                    + f"\n\nThis is likely due to hash changes in '{job.name}' node inputs. "
                    f"Current values and hashes: {job.inputs}, "
                    f"{job.task._hash}\n\n"
                    "Set loglevel to 'debug' in order to track hash changes "
                    "throughout the execution of the workflow.\n\n "
                    "These issues may have been caused by `bytes_repr()` methods "
                    "that don't return stable hash values for specific object "
                    "types across multiple processes (see bytes_repr() "
                    '"singledispatch "function in pydra/utils/hash.py).'
                    "You may need to write specific `bytes_repr()` "
                    "implementations (see `pydra.utils.hash.register_serializer`) or a "
                    "`__bytes_repr__()` dunder methods to handle one or more types in "
                    "your interface inputs."
                )
            val = res.get_output_field(self._field)
            val = self._apply_cast(val)
            return val

        # Get the execution node that the value is coming from
        upstream_node = graph.node(self._node.name)

        if not upstream_node._tasks:  # No jobs, return empty state array
            return StateArray()
        if not upstream_node.state:  # Return the singular job
            value = retrieve_from_job(upstream_node._tasks[None])
            if state_index is not None:
                return value[state_index]
            return value
        if upstream_node.state.combiner:

            # No state remains after the combination, return all values in a list
            if not upstream_node.state.ind_l_final:
                return [retrieve_from_job(j) for j in upstream_node.tasks]

            # Group the values of the tasks into list before returning
            def group_values(index: int) -> list:
                # Get a slice of the tasks that match the given index of the state array of the
                # combined values
                final_index = set(upstream_node.state.states_ind_final[index].items())
                return [
                    retrieve_from_job(upstream_node._tasks[i])
                    for i, ind in enumerate(upstream_node.state.states_ind)
                    if set(ind.items()).issuperset(final_index)
                ]

            if state_index is None:  # return all groups if no index is given
                return StateArray(
                    group_values(i) for i in range(len(upstream_node.state.ind_l_final))
                )
            return group_values(state_index)  # select the group that matches the index
        if state_index is None:  # return all jobs in a state array
            return StateArray(retrieve_from_job(j) for j in upstream_node.tasks)
        # Select the job that matches the index
        return retrieve_from_job(upstream_node._tasks[state_index])

    @property
    def _source(self):
        return self._node
